/************************************************************************************
* This program is licensed under the BSD two-clause license, a copy of which
* is included below.
*
* Copyright (c) 2012, Claudio D. Lapilli
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* x-Redistributions of source code must retain the above copyright notice, this list
*   of conditions and the following disclaimer.
* x-Redistributions in binary form must reproduce the above copyright notice, this
*   list of conditions and the following disclaimer in the documentation and/or other
*   materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
**************************************************************************************/





#include "cleanfs.h"

#define FSDEBUG     // ENABLE DEBUGGING

#include "implement.h"  // INCLUDE USER-IMPLEMENTED FUNCTIONS HERE.



clfs_64bit FSReadCluster(clfs_64bit cluster,clfs_byte *buffer,clfs_64bit nclusters,struct FS_VOLUME *fs)
{
    // WARNING: LOW-LEVEL ROUTINE, NO CHECK TO SEE IF READING BEYOND END OF DISK!
return    fs->Disk.read(fs->Disk.FirstSector+cluster*fs->SectorsPerCluster,buffer,nclusters*fs->SectorsPerCluster);
}
clfs_64bit FSWriteCluster(clfs_64bit cluster,clfs_byte *buffer,clfs_64bit nclusters,struct FS_VOLUME *fs)
{
    // WARNING: LOW-LEVEL ROUTINE, NO CHECK TO SEE IF READING BEYOND END OF DISK!
return    fs->Disk.write(fs->Disk.FirstSector+cluster*fs->SectorsPerCluster,buffer,nclusters*fs->SectorsPerCluster);
}

int FSInitFileBuffers(struct FS_FILE *file)
{
    // INITIALIZE WRITE BUFFER IF NEEDED
    if(!file->Buffer.Data) {
    // FSMalloc ONE CLUSTER READ BUFFER
    file->Buffer.Data=(clfs_byte *)FSMalloc(file->Volume->ClusterSize);
    if(!file->Buffer.Data) return FS_NOMEM; 	// NOT ENOUGH MEMORY FOR WRITE BUFFER, FAIL TO ALLOCATE CLUSTERS
    file->Buffer.Used=0;
    }
    return FS_OK;
}

int FSBufferFlush(struct FS_FILE *file)
{
    clfs_64bit byteswritten;
    clfs_64bit clusternum;

// CHECK IF ALREADY CACHED

    if(!file->Buffer.Data) return FS_OK;

    if(file->Buffer.Used&FSBUFFER_DIRTY) {
        // FLUSH USED BUFFER TO DISK BEFORE REUSING THE BUFFER
        clusternum=FSOffsetToCluster(file->Buffer.Offset,file);
        if(clusternum==-1) {
            // SOMETHING DAMAGED THE CLUSTER CHAIN ON THIS FILE
            return FS_ERROR;
        }

        byteswritten=FSWriteCluster(clusternum,file->Buffer.Data,1, file->Volume);
        if(byteswritten!=file->Volume->ClusterSize) return FS_CANTWRITE;
        file->Buffer.Used=FSBUFFER_HASDATA;
    }

    return FS_OK;
}


int FSBufferFill(struct FS_FILE *file,clfs_64bit NewOffset)
{
    clfs_64bit byteswritten;
    clfs_64bit clusternum;

    // CHECK IF OFFSET IS WITHIN FILE
    if(NewOffset<0) return FS_EOF;
    NewOffset&=(~(file->Volume->ClusterSize-1));
    if( NewOffset > file->F.FileSize) {
        return FS_EOF;
    }


    // CHECK IF ALREADY CACHED

    if(!((file->Buffer.Used) && (file->Buffer.Offset==NewOffset))) {
        int error;

        // FLUSH THE BUFFER IF IT CONTAINS DATA

        error=FSBufferFlush(file);
        if(error!=FS_OK) return error;


        // MUST READ THE CLUSTER INTO THE BUFFER FIRST
        clusternum=FSOffsetToCluster(NewOffset,file);
        if(clusternum==-1) {
            // TRYING TO READ PAST END-OF-FILE AND PAST END-OF-CHAIN
            return FS_EOF;
        }
        byteswritten=FSReadCluster(clusternum,file->Buffer.Data,1, file->Volume);
        if(byteswritten!=file->Volume->ClusterSize) {    file->Buffer.Used=0; return FS_CANTREAD; }

        file->Buffer.Used=FSBUFFER_HASDATA;
        file->Buffer.Offset=NewOffset;
   }

   return FS_OK;
}

inline void  FSBufferMarkDirty(struct FS_FILE *file)
{
    // CAREFUL, ONLY USE ON BUFFERS PROPERLY FILLED, DOES NOT CHECK IF FSBUFFER_HASDATA IS SET
    file->Buffer.Used|=FSBUFFER_DIRTY;
}



// MAP A CERTAIN Offset WITHIN THE READ BUFFERS.
clfs_byte *FSBufferMap(struct FS_FILE *file,clfs_64bit Offset)
{
    int error;

    error=FSBufferFill(file,Offset);
    if(error!=FS_OK) return NULL;

    return file->Buffer.Data+(Offset-file->Buffer.Offset);
}





// RETURN NUMBER OF CLUSTERS ALLOCATED IN A SINGLE FRAGMENT, UP TO needed CLUSTERS
// ALSO RETURN THE CLUSTER NUMBER IN *foundcluster
struct FS_RECORDLIST *FSAllocClusters(clfs_64bit requestedcluster,clfs_64bit needed,struct FS_VOLUME *fs)
{

    int error;
    clfs_32bit mask;
    clfs_32bit *ptr,*endofbuffer;
    clfs_64bit Offset,cluster;
    struct FS_RECORDLIST *frl;
    struct FS_FRAGRECORD *fr;

    if(!fs->FreeClusters) return NULL;  // NO FREE CLUSTER BITMAP AVAILABLE, CANNOT ALLOCATE
    if(requestedcluster==-1) return NULL; // ERROR CONDITION RETURNED FROM FSNextFreeCluster

    if(FSInitFileBuffers(fs->FreeClusters)!=FS_OK) return NULL;


    endofbuffer=(clfs_32bit *)(fs->FreeClusters->Buffer.Data+fs->ClusterSize);

    cluster=requestedcluster;

    mask=1<<(cluster&31);
    Offset=(cluster>>3)&(~(fs->ClusterSize-1));
    ptr=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
    ptr+=((cluster>>3)&(fs->ClusterSize-1))>>2;        // FIND PROPER WORD WITHIN BUFFER

    // READ CLUSTER
    error=FSBufferFill(fs->FreeClusters,Offset);
    if(error!=FS_OK) return NULL;

    // HAVE PROPER CLUSTER IN BUFFER

    if((*ptr)& mask) {
        // REQUESTED CLUSTER IS ALREADY USED
        // ALLOCATE CLUSTERS FROM THE NEXT FSFree AREA
        return FSAllocClusters(((requestedcluster==fs->NextFreeClusterHint)? FSNextFreeCluster(fs):fs->NextFreeClusterHint),needed,fs);
    }

    // REQUESTED CLUSTERS ARE AVAILABLE, GET AS MANY CLUSTERS AS POSSIBLE FOR FIRST FRAGMENT

    frl=(struct FS_RECORDLIST *)FSMalloc(sizeof(struct FS_RECORDLIST));
    if(!frl) return NULL; // CAN'T ALLOCATE ANY MEMORY

    FSMemset(frl,0,sizeof(struct FS_RECORDLIST));

    fr=&(frl->Fragrec);
    fr->EntryType=DE_FRAGMENT;
    fr->Fragments[0].StartCluster=cluster;
    fr->Fragments[0].NumClusters=0;

    while(needed) {
        if((*ptr)&mask) break;

        *ptr|=mask; // MARK CLUSTER AS TAKEN
        ++fr->Fragments[0].NumClusters;
        FSBufferMarkDirty(fs->FreeClusters);

        --needed;

        mask<<=1;
        if(!mask) {
            mask=1;
            ++ptr;
            if(ptr==endofbuffer) {
                // ADVANCE TO NEXT CLUSTER
                Offset+=fs->ClusterSize;
                error=FSBufferFill(fs->FreeClusters,Offset);
                if(error!=FS_OK) {
                    // ERROR READING THE BITMAP
                    FSFree(frl);
                    return NULL;
                }
                  ptr=(clfs_32bit *)fs->FreeClusters->Buffer.Data;    // POINT TO START OF NEW BUFFERR
            }


            }



        }

       // END OF LOOP: EITHER WE FOUND ALL THE CLUSTERS WE NEEDED OR WE NEED MORE

    if(needed) {
        frl->Next=FSAllocClusters(((requestedcluster==fs->NextFreeClusterHint)? FSNextFreeCluster(fs):fs->NextFreeClusterHint),needed,fs);
        if(!frl->Next) {
            FSFree(frl);
            return NULL;
        }
    }
    else  { frl->Next=NULL; if(requestedcluster==fs->NextFreeClusterHint) fs->NextFreeClusterHint=fr->Fragments[0].StartCluster+fr->Fragments[0].NumClusters; }    // SET NEXT HINT IMMEDIATELY AFTER LAST FRAGMENT
    return frl;

}

clfs_64bit FSGetLastCluster(struct FS_FILE *file)
{
    struct FS_RECORDLIST *fr=file->Chain;
    int f;
    // FIND LAST BLOCK
    if(!fr) return -1;

    while(fr->Next) fr=fr->Next;

    for(f=DIRENTRY_NUM_FRAGMENTS-1;f>=0;--f)
    {
        if(fr->Fragrec.Fragments[f].NumClusters) return fr->Fragrec.Fragments[f].StartCluster+fr->Fragrec.Fragments[f].NumClusters-1;
    }
    // ONLY IF FILE HAS NO STORAGE WHATSOEVER
    return -1;
}

clfs_64bit FSOffsetToCluster(clfs_64bit Offset,struct FS_FILE *file)
{
    struct FS_RECORDLIST *fr=file->Chain;
    int f;
    if(Offset<0) return -1;
    while(fr) {
    for(f=0;f<DIRENTRY_NUM_FRAGMENTS;++f)
    {
        if(fr->Fragrec.Fragments[f].NumClusters) {
            if(Offset<fr->Fragrec.Fragments[f].NumClusters*file->Volume->ClusterSize) {
                // OFFSET FALLS WITHIN THIS FRAGMENT
                return fr->Fragrec.Fragments[f].StartCluster+(Offset>>file->Volume->ClusterBits);
            }
            Offset-=fr->Fragrec.Fragments[f].NumClusters*file->Volume->ClusterSize;
        }
    }
        fr=fr->Next;

    }

    // OFFSET WAS OUT OF RANGE
    return -1;
}






void FSGetFragmentPtr(clfs_64bit Offset,struct FS_FILE *file,struct FS_RECORDLIST **frag,clfs_32bit *idx)
{
    struct FS_RECORDLIST *fr=file->Chain;
    int f;
    while(fr) {
    for(f=0;f<DIRENTRY_NUM_FRAGMENTS;++f)
    {
        if(fr->Fragrec.Fragments[f].NumClusters) {
            Offset-=fr->Fragrec.Fragments[f].NumClusters*file->Volume->ClusterSize;
            if(Offset<0) {
                *frag=fr;
                *idx=f;
                return;
            }
        }
    }
        fr=fr->Next;

    }

    *frag=NULL;
}

void FSGetNextFragment(struct FS_RECORDLIST **frag,clfs_32bit *idx)
{
    struct FS_RECORDLIST *fr=*frag;
    int f=*idx;
    while(fr) {

        ++f;
        if(f>=DIRENTRY_NUM_FRAGMENTS) {
            fr=fr->Next;
            f=-1;
            continue;
        }
        if(fr->Fragrec.Fragments[f].NumClusters) {
                *frag=fr;
                *idx=f;
                return;
        }
    }

    *frag=NULL;
}


// APPEND THE FRAGMENTS OF CLUSTERS IN RECORDLIST TO THE END OF THE FILE
// GIVEN FRAGMENTS ARE FREED AUTOMATICALLY OR ATTACHED DIRECTLY TO THE FILE CHAIN

void FSAppendFragments(struct FS_FILE *file,struct FS_RECORDLIST *fragments)
{
    clfs_64bit Num;
    struct FS_RECORDLIST *fr,*src;
    int fridx,srcidx;

    fr=file->Chain;

    if(!fr) {
        // FILE HAD NO STORAGE
        file->Chain=fragments;
    }
    else {

    // GO TO THE END OF CHAIN
    while(fr->Next) fr=fr->Next;

    // APPEND GIVEN FRAGMENTS
    fr->Next=fragments;

    }

    // COMPACT FRAGMENT CHAIN
    fr=file->Chain;
    fridx=0;

    src=fr;
    srcidx=fridx+1;

    do {

        // SCAN ALL FRAGMENTS WITH src
        // IGNORING EMPTY FRAGMENTS
        if(src->Fragrec.Fragments[srcidx].NumClusters==0) {
            ++srcidx;
            if(srcidx>=DIRENTRY_NUM_FRAGMENTS) {
                src=src->Next;
                if(!src) break;
                srcidx=0;
            }
            continue;
        }

        // HERE, fr POINTS TO A VALID FRAGMENT, AND src POINTS TO A CANDIDATE FOR MERGER


        if(fr->Fragrec.Fragments[fridx].NumClusters) {
        // CHECK IF FRAGMENTS ARE CONTIGUOUS
        if(src->Fragrec.Fragments[srcidx].StartCluster==fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters)
        {
            // MERGE BOTH FRAGMENTS, BUT ONLY IF THERE'S NO OVERFLOW IN NumClusters
            Num=(clfs_64bit)fr->Fragrec.Fragments[fridx].NumClusters+(clfs_64bit)src->Fragrec.Fragments[srcidx].NumClusters;
            if(Num>>32) {
                // ADVANCE fr TO NEXT SLOT, IT CAN EITHER BE SAME AS src OR BE AN EMPTY SLOT
                ++fridx;
                if(fridx>=DIRENTRY_NUM_FRAGMENTS) {
                    fr=fr->Next;
                    if(!fr) break;
                    fridx=0;
                }
            }
            else {
                // MERGE BOTH FRAGMENTS
            fr->Fragrec.Fragments[fridx].NumClusters+=src->Fragrec.Fragments[srcidx].NumClusters;
            src->Fragrec.Fragments[srcidx].NumClusters=0;

            // ADVANCE src AND CONTINUE
                ++srcidx;
                if(srcidx>=DIRENTRY_NUM_FRAGMENTS) {
                    src=src->Next;
                    if(!src) break;
                    srcidx=0;
                }


            continue;
            }
        }
        else {
            // FRAGMENTS CANNOT BE MERGED, JUST MOVE src RIGHT AFTER fr AND USE IT FOR NEXT ITERATION
            // ADVANCE fr TO NEXT SLOT, IT CAN EITHER BE SAME AS src OR BE AN EMPTY SLOT
            ++fridx;
            if(fridx>=DIRENTRY_NUM_FRAGMENTS) {
                fr=fr->Next;
                if(!fr) break;
                fridx=0;
            }


        }
        } // ELSE THIS IS AN EMPTY FRAGMENT AND SHOULD BE OVERWRITTEN WITH src

        // MOVE src FRAGMENT TO fr IF THEY ARE DIFFERENT
        if( !((fr==src)&&(fridx==srcidx)) ) {
        fr->Fragrec.Fragments[fridx].StartCluster=src->Fragrec.Fragments[srcidx].StartCluster;
        fr->Fragrec.Fragments[fridx].NumClusters=src->Fragrec.Fragments[srcidx].NumClusters;
        src->Fragrec.Fragments[srcidx].NumClusters=0;
        }

        // ADVANCE src AND CONTINUE
            ++srcidx;
            if(srcidx>=DIRENTRY_NUM_FRAGMENTS) {
                src=src->Next;
                if(!src) break;
                srcidx=0;
            }


    } while(fr);

    // REMOVE AND FREE ANY EMPTY SLOTS

    fr=file->Chain;
    src=NULL;

    while(fr) {

        for(fridx=0;fridx<DIRENTRY_NUM_FRAGMENTS;++fridx)
        {
            if(fr->Fragrec.Fragments[fridx].NumClusters) break;
        }
        if(fridx==DIRENTRY_NUM_FRAGMENTS) {
            // REMOVE THIS EMPTY BLOCK
            if(src) src->Next=fr->Next;
            else file->Chain=fr->Next;
            FSFree(fr);
            if(src) fr=src->Next;
            else fr=file->Chain;
        }
        else {
            src=fr;
            fr=fr->Next;
        }

    }

}

// APPEND THE METADATA IN RECORDLIST TO THE END OF THE EXISTING METADATA
// GIVEN RECORDS ARE FREED AUTOMATICALLY OR ATTACHED DIRECTLY TO THE FILE CHAIN

void FSAppendMetadata(struct FS_FILE *file,struct FS_RECORDLIST *meta)
{
    struct FS_RECORDLIST *fr;

    fr=file->Metadata;

    if(!fr) {
        // FILE HAD NO STORAGE
        file->Metadata=meta;
    }
    else {

    // GO TO THE END OF CHAIN
    while(fr->Next) fr=fr->Next;

    // APPEND GIVEN FRAGMENTS
    fr->Next=meta;

    }


}




// GROW A FILE TO A CERTAIN SIZE BY ALLOCATING CLUSTERS TO IT
// WARNING: THIS OPERATION SHOULD BE ATOMIC, LOCK FILESYSTEM UNTIL FINISHED

int FSGrowFile(clfs_64bit filesize,struct FS_FILE *file)
{
    clfs_64bit clustercount=0,needed;
    struct FS_RECORDLIST *fr,*newfr;
    struct FS_VOLUME *fs=file->Volume;
    int f;

    fr=file->Chain;

    // DETERMINE HOW MANY CLUSTERS ARE ALREADY ALLOCATED
    while(fr) {
        for(f=0;f<DIRENTRY_NUM_FRAGMENTS;++f)
        {
            clustercount+=fr->Fragrec.Fragments[f].NumClusters;
        }
    fr=fr->Next;
    }


    // NUMBER OF NEW CLUSTERS TO ALLOCATE
    needed=((filesize + fs->ClusterSize-1) )>>fs->ClusterBits;
    needed-=clustercount;

    if(needed>0) {

    // ATTEMPT TO ALLOC CLUSTERS FROM PREFERRED LOCATION
    if(clustercount) newfr=FSAllocClusters(FSGetLastCluster(file)+1,needed,fs);
    else newfr=FSAllocClusters(fs->NextFreeClusterHint,needed,fs);
        if(newfr) {
                FSAppendFragments(file,newfr);
        }
        else return FS_DISKFULL;    // UNABLE TO RETRIEVE ENOUGH CLUSTERS

    }
    return FS_OK;

}





// SAME AS FSREADLL BUT WITH ARGUMENT CHECKING
clfs_64bit FSRead(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file)
{
    if((!file) || (!buffer)) return 0;
    if(nbytes<0) return 0;
    return FSReadLL(buffer,nbytes,file);
}

// LOW-LEVEL READ FROM FILE'S CURRENT POSITION
// WITH INTERNAL BUFFERING

clfs_64bit FSReadLL(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file)
{
clfs_64bit currentclstr,bytesread,totalcount,bytesstart,bytesend;
clfs_64bit nclusters;
struct FS_RECORDLIST *fr;
int fridx;
int error;

totalcount=0;


// RETURN IF EOF
if(file->CurrentOffset>=file->F.FileSize) { return totalcount; }






// TRUNCATE IF TRYING TO READ BEYOND EOF
if(file->CurrentOffset+nbytes>file->F.FileSize) nbytes=file->F.FileSize-file->CurrentOffset;

// READ IN 3 PARTS: START, FULL-CLUSTERS, AND END


// FIRST PART: START
while(nbytes) {

bytesstart=file->CurrentOffset&(file->Volume->ClusterSize-1);

if(bytesstart || (nbytes<file->Volume->ClusterSize)) {
    // FIRST/LAST CLUSTER NEEDS TO BE READ PARTIALLY
    // INITIALIZE READ BUFFER IF NEEDED

    // INITIALIZE READ BUFFERS
    error=FSInitFileBuffers(file);
    if(error!=FS_OK) return totalcount;

    error=FSBufferFill(file,file->CurrentOffset&~(file->Volume->ClusterSize-1));
    if(error!=FS_OK) return totalcount;

    // DATA IS IN BUFFER, COPY TO FINAL
    bytesend=bytesstart+nbytes;
    if(bytesend>file->Volume->ClusterSize) bytesend=file->Volume->ClusterSize;  // TRIM TO 1 CLUSTER

    FSMemcpy(buffer,file->Buffer.Data+bytesstart,bytesend-bytesstart);
    nbytes-=bytesend-bytesstart;
    totalcount+=bytesend-bytesstart;
    file->CurrentOffset+=bytesend-bytesstart;
    buffer+=bytesend-bytesstart;

}

// DONE WITH INITIAL READ, NOW READ FULL CLUSTERS


// START READING FULL FRAGMENTS

nclusters=nbytes>>file->Volume->ClusterBits;   // TOTAL FULL CLUSTERS TO READ
if(nclusters) {
FSGetFragmentPtr(file->CurrentOffset,file,&fr,&fridx);
currentclstr=FSOffsetToCluster(file->CurrentOffset,file);
if(currentclstr==-1) {
    // SOMETHING IS VERY WRONG WITH THIS FILE'S ALLOCATION CHAIN
    return totalcount;
}
while(nclusters+currentclstr>=fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters) {

// READ ENTIRE FRAGMENT DIRECTLY INTO FINAL BUFFER
bytesread=FSReadCluster(currentclstr,buffer,fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters-currentclstr,file->Volume);
totalcount+=bytesread;
file->CurrentOffset+=bytesread;
if(bytesread!=(fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters-currentclstr)*file->Volume->ClusterSize) {
// ERROR READING LAST BLOCK, RETURN WHAT WAS READ SO FAR
return totalcount;
}
nbytes-=bytesread;
nclusters-=bytesread>>file->Volume->ClusterBits;
buffer+=bytesread;

FSGetNextFragment(&fr,&fridx);
if(fr==NULL) return totalcount;
currentclstr=fr->Fragrec.Fragments[fridx].StartCluster;
}
}

// FINISH READING FULL CLUSTERS IN A PARTIAL FRAGMENT
if(nclusters) {
// READ FULL CLUSTERS DIRECTLY INTO BUFFER
bytesread=FSReadCluster(currentclstr,buffer,nclusters,file->Volume);
totalcount+=bytesread;
file->CurrentOffset+=bytesread;
currentclstr+=bytesread>>file->Volume->ClusterBits;
buffer+=bytesread;
nbytes-=bytesread;
if(bytesread!=nclusters<<file->Volume->ClusterBits) return totalcount;  // ERROR, INCOMPLETE READING
nclusters=0;
}

}

return totalcount;

}


// SAME AS FSWRITELL WITH SOME ARGUMENT CHECKING
clfs_64bit FSWrite(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file)
{
    if((!file) || (!buffer)) return 0;
    if(nbytes<0) return 0;
    return FSWriteLL(buffer,nbytes,file);
}

clfs_64bit FSWriteLL(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file)
{
clfs_64bit currentclstr,bytesread,totalcount,bytesstart,bytesend;
clfs_64bit nclusters;
struct FS_RECORDLIST *fr;
clfs_32bit fridx;
int error;

totalcount=0;

// RETURN IF FILE NOT OPEN FOR WRITING
if(!(file->Mode&FSMODE_WRITE)) return 0;

// MARK FILE TO UPDATE ENTRY
file->Mode|=FSMODE_DIRTYDIR;

// DON'T ALLOW OVERWRITING WHEN IN APPEND MODE
if(file->Mode&FSMODE_APPEND) file->CurrentOffset=file->F.FileSize;

// CHECK IF WE NEED TO GROW THE FILE
if(file->CurrentOffset+nbytes>file->F.FileSize) {
    // GROW THE CLUSTER CHAIN
    if(file->Mode&FSMODE_NOGROW) {
        // TRUNCATE nbytes
        nbytes=file->F.FileSize-file->CurrentOffset;
        if(nbytes<=0) return totalcount;
    }
    else if(FSGrowFile(file->CurrentOffset+nbytes,file)!=FS_OK) return 0;
}


// WRITE IN 3 PARTS: START, FULL-CLUSTERS, AND END


// FIRST PART: START
while(nbytes) {

bytesstart=file->CurrentOffset&(file->Volume->ClusterSize-1);

if(bytesstart || (nbytes<file->Volume->ClusterSize)) {
    // FIRST/LAST CLUSTER NEEDS TO BE READ PARTIALLY
    // INITIALIZE READ BUFFER IF NEEDED

    // INITIALIZE READ BUFFERS
    error=FSInitFileBuffers(file);
    if(error!=FS_OK) return totalcount;

    error=FSBufferFill(file,file->CurrentOffset&~(file->Volume->ClusterSize-1));
    if(error!=FS_OK) return totalcount;

    // DATA IS IN BUFFER, COPY TO FINAL
    bytesend=bytesstart+nbytes;
    if(bytesend>file->Volume->ClusterSize) bytesend=file->Volume->ClusterSize;  // TRIM TO 1 CLUSTER

    FSMemcpy(file->Buffer.Data+bytesstart,buffer,bytesend-bytesstart);
    FSBufferMarkDirty(file);
    nbytes-=bytesend-bytesstart;
    totalcount+=bytesend-bytesstart;
    file->CurrentOffset+=bytesend-bytesstart;
    if(file->CurrentOffset>file->F.FileSize) file->F.FileSize=file->CurrentOffset;
    buffer+=bytesend-bytesstart;

}

// DONE WITH INITIAL WRITE, NOW WRITE FULL CLUSTERS


// START WRITING FULL FRAGMENTS

nclusters=nbytes>>file->Volume->ClusterBits;   // TOTAL FULL CLUSTERS TO READ
if(nclusters) {
FSGetFragmentPtr(file->CurrentOffset,file,&fr,&fridx);
currentclstr=FSOffsetToCluster(file->CurrentOffset,file);
if(currentclstr==-1) {
    // FILE HAS A DAMAGED OR TRUNCATED ALLOCATION CHAIN
    return totalcount;

}
while(nclusters+currentclstr>=fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters) {

// WRITE ENTIRE FRAGMENT DIRECTLY INTO FINAL BUFFER
bytesread=FSWriteCluster(currentclstr,buffer,fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters-currentclstr,file->Volume);
totalcount+=bytesread;
file->CurrentOffset+=bytesread;
if(file->CurrentOffset>file->F.FileSize) file->F.FileSize=file->CurrentOffset;
if(bytesread!=(fr->Fragrec.Fragments[fridx].StartCluster+fr->Fragrec.Fragments[fridx].NumClusters-currentclstr)*file->Volume->ClusterSize) {
// ERROR WRITING LAST BLOCK, RETURN WHAT WAS WRITTEN SO FAR
return totalcount;
}
nbytes-=bytesread;
nclusters-=bytesread>>file->Volume->ClusterBits;
buffer+=bytesread;

FSGetNextFragment(&fr,&fridx);
if(fr==NULL) return totalcount;
currentclstr=fr->Fragrec.Fragments[fridx].StartCluster;
}
}
// FINISH READING FULL CLUSTERS IN A PARTIAL FRAGMENT
if(nclusters) {
// READ FULL CLUSTERS DIRECTLY INTO BUFFER
bytesread=FSWriteCluster(currentclstr,buffer,nclusters,file->Volume);
totalcount+=bytesread;
file->CurrentOffset+=bytesread;
if(file->CurrentOffset>file->F.FileSize) file->F.FileSize=file->CurrentOffset;
currentclstr+=bytesread>>file->Volume->ClusterBits;
buffer+=bytesread;
nbytes-=bytesread;
if(bytesread!=nclusters<<file->Volume->ClusterBits) return totalcount;  // ERROR, INCOMPLETE READING
nclusters=0;
}

}

if(!(file->Mode&FSMODE_MODIFY)) file->F.FileSize=file->CurrentOffset;   // TRUNCATE FILE UNLESS THE MODIFY FLAG WAS USED

return totalcount;

}


// INITIALIZE VOLUME, MOUNT IT

int FSMountVolume(struct FS_DISK *Disk,struct FS_VOLUME *fs)
{
// STRUCTURE FS_DISK SHOULD BE FILLED BY THE USER

    clfs_byte *sectorbuffer=(clfs_byte *)FSMalloc(Disk->SectorSize);
    if(!sectorbuffer) return FS_NOMEM;

    // READ FIRST SECTOR
    clfs_64bit bytesread;

    struct FS_BOOTRECORD *boot;

    // LOOK FOR BOOT RECORD WITHIN THE FIRST 64 SECTORS
    clfs_32bit sectors=0;
    while(sectors<64) {
        bytesread=Disk->read(Disk->FirstSector+sectors,sectorbuffer,1);
        if(bytesread!=Disk->SectorSize) { ++sectors; continue; }

        boot=(struct FS_BOOTRECORD *)sectorbuffer;

        while(((clfs_byte *)boot)<sectorbuffer+Disk->SectorSize) {

            if((boot->EntryType==DE_BOOTRECORD)&&(boot->Magic==FSBOOTRECORD_MAGIC)) break;
            ++boot;
            }
        if(((clfs_byte *)boot)<sectorbuffer+Disk->SectorSize) break;

        ++sectors;

        }

    if(sectors==64) {
        // BOOT RECORD NOT FOUND
        FSFree(sectorbuffer);
        return FS_INVALIDDATA;
    }


    // BOOT RECORD FOUND,GET INFORMATION FOR THE VOLUME

    FSMemset(fs,0,sizeof(struct FS_VOLUME));
    FSMemcpy(&fs->Disk,Disk,sizeof(struct FS_DISK));
    fs->ClusterSize=boot->BytesPerCluster;
    fs->ClusterBits=1;
    while( (1<<fs->ClusterBits)<fs->ClusterSize) ++fs->ClusterBits;
    fs->NumClusters=boot->TotalClusters;
    fs->SectorsPerCluster=boot->BytesPerCluster/Disk->SectorSize;
    // LEAVE Flags = 0 FOR NOW

    // EXTRACT VOLUME NAME
    struct FS_NAMERECORD *namer=(struct FS_NAMERECORD *)(boot+1);
    clfs_32bit nnameentries=0,offset=((clfs_byte *)boot)-sectorbuffer;

    // COUNT NUMBER OF NAME ENTRIES, LIMIT TO ONE SECTOR
    while(namer->EntryType==DE_NAME) {
        ++nnameentries;
        namer++;
        offset+=sizeof(struct FS_NAMERECORD);
        if(offset>=bytesread) break;      // STOP AT END OF BOOT SECTOR
    }

    fs->VolName=(clfs_char *)FSMalloc(BOOTRECORD_NAME_LENGTH+NAMERECORD_LENGTH*nnameentries+1);
    if(fs->VolName==NULL) {
        FSFree(sectorbuffer);
        return FS_NOMEM;
    }

    FSMemcpy(fs->VolName,boot->VolName,BOOTRECORD_NAME_LENGTH);
    namer=(struct FS_NAMERECORD *)(boot+1);
    offset=DIRENTRY_NAME_LENGTH;
    while(nnameentries) {
        FSMemcpy(fs->VolName+offset,namer->Name,NAMERECORD_LENGTH);
        --nnameentries;
        offset+=NAMERECORD_LENGTH;
    }
    fs->VolName[offset]=0;  // ADD NULL-TERMINATOR FOR SPECIAL CASE


    // NOW MOUNT AND OPEN THE ROOT DIRECTORY
    // USE LOW-LEVEL FUNCTION
    int error=FSOpenDirLL(&(fs->RootDir),fs,boot->RootDirCluster);

    if(error!=FS_OK) {
        FSFree(fs->VolName);
        fs->VolName=NULL;
        FSFree(sectorbuffer);
        return error;
    }

    // SET CURRENT DIRECTORY AT ROOT
    fs->CurrentDir=&(fs->RootDir);

    fs->Flags|=FSATTR_IMMUTABLE;

    FSFree(sectorbuffer);

    // NOW FIND FREE SPACE BITMAP IF IT EXISTS

    error=FSMountFreeSpace(fs);

    if(error!=FS_OK) {
        if(error==FS_NOTFOUND) {
            // FREE SPACE BITMAP NEEDS TO BE CREATED BY SPECIAL TOOLS
            fs->Flags|=FSATTR_IMMUTABLE;
            return FS_OK;
        }
        if(error==FS_INVALIDDATA) {
            // DISK WASN'T UNMOUNTED PROPERLY LAST TIME, ATTEMPT TO REBUILD

            error=FSRescanFreeSpace(fs);
            if(error!=FS_OK) {
                fs->Flags|=FSATTR_IMMUTABLE;
                return FS_OK;
            }
            return FS_OK;


        }
        // ANY OTHER ERROR CAN BE IGNORED, BUT MOUNT READ-ONLY
        fs->Flags|=FSATTR_IMMUTABLE;
        return FS_OK;
    }

    return FS_OK;

}


// AUXILIARY FILENAME PARSING FUNCTIONS

clfs_32bit __fsstrlen(clfs_char *string)
{
    clfs_32bit count=0;
    while(string[count]) ++count;
    return count;
}


clfs_char *__fsfindchar(clfs_char *strstart,clfs_char *strend,const clfs_char *chars)
{
char a,*ptr;
while(strstart!=strend)
{
a=*strstart;
if(!a) return NULL;
ptr=(char *)chars;
while(*ptr!=0) { if(a==*ptr) return strstart; ++ptr; }
++strstart;
}
return NULL;
}


clfs_char *__fsfindcharrev(clfs_char *strstart,clfs_char *strend,const clfs_char *chars)
{
clfs_char a,*ptr;
if(strend==NULL) {
strend=strstart;
while(*strend!=0) ++strend;
}
while(strstart!=strend)
{
--strend;
a=*strend;
ptr=(clfs_char *)chars;
while(*ptr!=0) { if(a==*ptr) return strend; ++ptr; }
}
return NULL;
}


// UNICODE UTF-8 HELPER FUNCTIONS
// RETURNS THE 32-BIT CODE FOR *ptr AND ADVANCES ptr TO THE NEXT CHAR
clfs_32bit __fsdecodeUTF8(clfs_char **ptr)
{
    clfs_char *tmp=*ptr;
    // ONE-BYTE CODE
    if(*tmp<0x80) { *ptr+=1; return *tmp; }
    if((*tmp&0xf8)==0xf0) {
        // 4-BYTE ENCODING
        *ptr=*ptr+4;
        return (((tmp[0]&0x7)<<18) | ((tmp[1] & 0x3f)<<12) | ((tmp[2] & 0x3f)<<6) | (tmp[3]& 0x3f));
    }
    if((*tmp&0xf0)==0xe0) {
        // 3-BYTE ENCODING
        *ptr=*ptr+3;
        return ((tmp[0]&0xF)<<12) | ((tmp[1]&0x3f)<<6) | ((tmp[2]&0x3f));
    }
    if((*tmp&0xe0)==0xc0) {
        // 2-BYTE ENCODING
        *ptr=*ptr+2;
        return ((tmp[0]&0x1f)<<6) | (tmp[1]&0x3f) ;
    }
    if((**ptr&0xc0)==0x80) { (*ptr)++; return -1; } // PARTIAL BYTE, INVALID CHARACTER

    return -1;
}



// RETURN FALSE IF THERE'S A MALFORMED UTF-8 BYTE SEQUENCE
int __fsisvalidUTF8(clfs_char *strstart)
{
    clfs_char a;
    while(1)
    {
    a=*strstart++;
    if(!a) return FS_TRUE;

    if(a<0x80) continue;


    // UNICODE SPEC - TABLE 3-7 CASES
    if(a<0xc2) return FS_FALSE;    // INVALID RANGE
    if(a<0xe0) {
        // TWO BYTE SEQUENCE, CHECK VALIDITY OF SECOND BYTE
        // U+0080 TO U+07FF
        if( (*strstart&0xc0) != 0x80) return FS_FALSE; // INVALID

        ++strstart;
        continue;

    }
    if(a<0xf0) {
        // 3-BYTE SEQUENCES
        // U+0800 to U+FFFF
        if( (a==0xe0) && (*strstart<0xa0)) return FS_FALSE; // MALFORMED (U+0000 TO U+07FF SHOULD BE A TWO-BYTE SEQUENCE)
        if( (a==0xed) && (*strstart>0x9f)) return FS_FALSE; // MALFORMED (U+D800 TO U+DFFF IS AN INVALID RANGE)
        if( (strstart[0]&0xc0) != 0x80) return FS_FALSE; // INVALID
        if( (strstart[1]&0xc0) != 0x80) return FS_FALSE; // INVALID
        strstart+=2;
        continue;
    }
    if(a<0xf5) {
        // 4-BYTE SEQUENCES
        // U+010000 to U+10FFFF
        if( (a==0xf0) && (*strstart<0x90)) return FS_FALSE; // MALFORMED (U+0000 TO U+FFFF SHOULD BE A 2 OR THREE-BYTE SEQUENCE)
        if( (a==0xf4) && (*strstart>0x8f)) return FS_FALSE; // MALFORMED ( > U+10FFFF IS INVALID)
        if( (strstart[0]&0xc0) != 0x80) return FS_FALSE; // INVALID
        if( (strstart[1]&0xc0) != 0x80) return FS_FALSE; // INVALID
        if( (strstart[2]&0xc0) != 0x80) return FS_FALSE; // INVALID
        strstart+=3;
        continue;
    }

    // ANYTHING ABOVE 0xF4 IS AN INVALID CHARACTERR
    return FS_FALSE;

    } // END WHILE

    // NO INVALID CHARACTERS DETECTED SO FAR, SO RETURN TRUE
    return FS_TRUE;

}

// RETURN TRUE IF THERE ARE STRANGE CONTROL CHARACTERS IN THE NAME
int __fshasinvalidcharsUTF8(clfs_char *strstart)
{
    clfs_char a;
    clfs_32bit chr;
    while(1)
    {
    a=*strstart++;
    if(!a) return FS_FALSE;

    // NO CONTROL CHARACTERS ALLOWED
    if(a<0x20) return FS_TRUE;

    if(a<0x7F) continue;

    // UNICODE SPEC - TABLE 3-7 CASES
    if(a<0xc2) return FS_TRUE;    // INVALID BYTES: THIS WILL ALSO DETECT 0X7F AS A CONTROL CHARACTER
    if(a<0xe0) {
        // 2-BYTE SEQUENCE, CHECK FOR SPECIAL INVALID CONTROL CHARACTERS
        chr=((a&0x1f)<<6) | (strstart[0]&0x3f);

        if(chr<0xA0) return FS_TRUE;        // (ALL ISO6429 CONTROL CHARACTERS ALSO INVALID)
        ++strstart;
        continue;

    }
    if(a<0xf0) {
        // 3-BYTE SEQUENCES
        // U+0800 to U+FFFF
        // 3-BYTE SEQUENCE, CHECK FOR SPECIAL INVALID CONTROL CHARACTERS
        chr=((a&0xf)<<6) | (strstart[0]&0x3f);
        chr=(chr<<6) | (strstart[1]&0x3f);

        //0x2028 and 0x2029 ARE LINE AND PARAGRAPH SEPARATORS, NOT ALLOWED IN A FILE NAME
        if(chr==0x2028) return FS_TRUE;
        if(chr==0x2029) return FS_TRUE;
        if(chr==0xfeff) return FS_TRUE; // BYTE ORDER MARK SHOULDN'T BE IN A FILE NAME
        if(chr>=0xfff0) return FS_TRUE; // ALL SPECIAL CHARACTERS
        if( (chr>=0xfdd0)&&(chr<=0xfdef) ) return FS_TRUE; // THESE ARE NON-CHARACTERS, NOT ALLOWED IN A NAME

        strstart+=2;
        continue;
    }

    // 4-BYTE SEQUENCES
    chr=((a&0xf)<<6) | (strstart[0]&0x3f);
    chr=(chr<<6) | (strstart[1]&0x3f);
    chr=(chr<<6) | (strstart[2]&0x3f);

    if( (chr&0xfffe)==0xfffe) return FS_TRUE; // DETECT 0XnnFFFE AND 0XnnFFFF IN ALL PLANES

    } // END WHILE

    // NO INVALID CHARACTERS DETECTED SO FAR, SO RETURN TRUE
    return FS_TRUE;

}


clfs_32bit __fscasefold(clfs_32bit character)
{
// THESE TABLES ARE FOR PROPER UNICODE CASE-INSENSITIVE COMPARISON
// TABLES ADD ABOUT 1400 BYTES TO LIBRARY
#include "folding_table.h"
    int idx;

    if(character<0x10000) {
        for(idx=0;folding_table16[idx].start!=0;++idx)
        {
            if(character<folding_table16[idx].start) return character;
            if(character<=folding_table16[idx].end) {
                if(folding_table16[idx].diff==1) {
                    if( (character-folding_table16[idx].start)&1) return character;
                }
                return character+folding_table16[idx].diff;
            }
    }
        return character;
    }
    for(idx=0;folding_table32[idx].start!=0;++idx)
    {
        if(character<folding_table32[idx].start) return character;
        if(character<=folding_table32[idx].end) {
            if(folding_table32[idx].diff==1) {
                if( (character-folding_table32[idx].start)&1) return character;
            }
            return character+folding_table32[idx].diff;
        }
    }

    return character;
}

const clfs_u32bit __crctable[256] =
{
 0, 0x77073096, 0xEE0E612C, 0x990951BA,
 0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3,
 0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988,
 0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91,
 0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE,
 0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7,
 0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC,
 0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5,
 0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172,
 0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B,
 0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940,
 0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59,
 0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116,
 0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F,
 0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924,
 0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D,
 0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A,
 0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433,
 0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818,
 0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01,
 0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E,
 0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457,
 0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C,
 0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65,
 0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2,
 0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB,
 0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0,
 0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9,
 0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086,
 0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F,
 0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4,
 0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD,
 0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A,
 0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683,
 0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8,
 0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1,
 0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE,
 0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7,
 0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC,
 0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5,
 0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252,
 0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B,
 0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60,
 0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79,
 0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236,
 0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F,
 0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04,
 0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D,
 0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A,
 0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713,
 0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38,
 0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21,
 0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E,
 0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777,
 0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C,
 0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45,
 0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2,
 0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB,
 0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0,
 0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9,
 0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6,
 0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF,
 0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94,
 0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D
};

// CALCULATE THE UNFOLDED NAME CRC

clfs_u32bit FSCalcNameCRC(clfs_char * data)
{
    clfs_u32bit crc=-1;
    while(*data) crc=__crctable[(crc ^ *data++) & 0xFF] ^ (crc >> 8);
    return crc^(-1);
}


clfs_u32bit FSCalcFoldedNameCRC(clfs_char * data)
{

    // WARNING: THIS CODE MIGHT CRASH WITH MALFORMED UTF-8 STRINGS

    clfs_char *data2=data;
    clfs_u32bit crc=-1;
    clfs_32bit unicodechar;
    while(*data2) {
        unicodechar=__fscasefold(__fsdecodeUTF8(&data2));
        if(unicodechar<0x7f)    // 1-BYTE ENCODING
        {
            crc=__crctable[(crc ^ unicodechar) & 0xFF] ^ (crc >> 8);
            continue;
        }
        if(unicodechar<0x800)    // 2-BYTE ENCODING
        {
            crc=__crctable[(crc ^ ((unicodechar>>6)|0xc0)) & 0xFF] ^ (crc >> 8);
            crc=__crctable[(crc ^ ((unicodechar&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
            continue;
        }
        if(unicodechar<0x10000)    // 3-BYTE ENCODING
        {
            crc=__crctable[(crc ^ ((unicodechar>>12)|0xe0)) & 0xFF] ^ (crc >> 8);
            crc=__crctable[(crc ^ (((unicodechar>>6)&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
            crc=__crctable[(crc ^ ((unicodechar&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
            continue;
        }
            // 4-BYTE ENCODING
        crc=__crctable[(crc ^ ((unicodechar>>18)|0xf0)) & 0xFF] ^ (crc >> 8);
        crc=__crctable[(crc ^ (((unicodechar>>12)&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
        crc=__crctable[(crc ^ (((unicodechar>>6)&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
        crc=__crctable[(crc ^ ((unicodechar&0x3f)|0x80)) & 0xFF] ^ (crc >> 8);
            continue;
    }
    return crc^(-1);
}




// COMPARE THE GIVEN NAME WITH THE NAME IN THE FS_FILERECORD STRUCTURE.
// COMPARE UP TO DIRENTRY_NAME_LENGTH CHARACTERS, RETURN TRUE IF THERE'S A MATCH, FALSE OTHERWISE
// STORE A POINTER IN nameptr, POINTING INSIDE THE ORIGINAL FILE NAME (name)
// TO THE POINT WHERE THERE WAS A MISMATCH, TO THE NULL TERMINATOR IF STRING IS LESS THAN DIRENTRY_NAME_LENGTH CHARACTERS
// OR TO THE NEXT CHARACTER TO BE COMPARED IF STRING IS LONGER

int FSPartialNameCompare(clfs_char *name,clfs_char *entryName,clfs_32bit *CharacterCache,clfs_char **nameptr,clfs_32bit entrychars,int CaseInsens)
{

    if(CaseInsens) {
    // PERFORM CASE INSENSITIVE COMPARISON
        // MUST DECODE UNICODE CHARS PROPERLY, THEN DO CASE-FOLDING
        // ALWAYS ASSUME VALID UTF-8 STRINGS, WILL FAIL THE COMPARISON WITH INVALID STRINGS
    int f;
    clfs_32bit partial,partcnt;
    partcnt=(*CharacterCache>>24)&0xff;
    partial=*CharacterCache&0xffffff;

      for(f=0;f<entrychars;++f)
      {
          if(partcnt) {
              if( ((*entryName)&0xc0)!=0x80) { *nameptr=name; return 0; }    // STRING IS NOT VALID UTF-8!
              partial=(partial<<6)|((*entryName)&0x3f);
              --partcnt;
              if(!partcnt) {
                  // GOT COMPLETE CHARACTER, COMPARE
                  *nameptr=name;
                  if(__fscasefold(partial)!=__fscasefold(__fsdecodeUTF8(nameptr))) { *nameptr=name; return 0; }   // STRINGS ARE DIFFERENT
                  name=*nameptr; // ADVANCE TO NEXT CHARACTER, SET BY __fsdecodeUTF8
              }
              ++entryName;
              continue;
          }
          if(*entryName&0x80) {
                        if(*name==0) { *nameptr=name; return 0; }   // QUIT NOW IF STRING WAS SUPPOSED TO END
           // DECODE UNICODE (PARTIAL)
              if( (*entryName&0xf8)==0xf0) { partcnt=3; partial=*entryName&0x7; }
              else {
                  if((*entryName&0xf0)==0xe0) { partcnt=2; partial=*entryName&0xf; }
                  else {
                      if((*entryName&0xe0)==0xc0) { partcnt=1; partial=*entryName&0x1f; }
                      else { *nameptr=name; return 0; } // INVALID UTF-8 STRING!
                  }
              }
          }
          else {
          if(*name!=*entryName) { *nameptr=name; return 0; }    // STRINGS ARE DIFFERENT, RETURN IMMEDIATELY
          if(*name==0) break;
          ++name;
          }
          ++entryName;

      }
      *CharacterCache=partial | (partcnt<<24);
      *nameptr=name;
      return 1;

}
else {
      int f;
        // PERFORM CASE SENSITIVE COMPARISON = BYTE-BY-BYTE EQUALITY
        for(f=0;f<entrychars;++f)
        {
            if(*name!=*entryName) { *nameptr=name; return 0; }    // STRINGS ARE DIFFERENT, RETURN IMMEDIATELY
            if(*name==0) break;
            ++name;
            ++entryName;
        }
        *nameptr=name;
        return 1;

    }

}


int FSNameCompare(clfs_char *name,clfs_char *othername,int CaseInsens)
{
clfs_32bit namelen=__fsstrlen(name);
clfs_32bit unnecessary_cache=0;
clfs_char *unnecessary_ptr;

// CALL PARTIAL NAME COMPARISON ROUTINE, IN THIS CASE WE DON'T NEED THE PARTIAL CACHE OR THE POSITION OF THE
// LAST CHARACTER THAT MATCHED, HENCE THE unnecessary PREFIX

return FSPartialNameCompare(name,othername,&unnecessary_cache,&unnecessary_ptr,namelen,CaseInsens);

}





// FIND A DIRECTORY ENTRY WITHIN A DIRECTORY AND FILL THE FS_FILE STRUCTURE
// NO WILDCARDS ACCEPTED. ONLY EXACT MATCH IS RETURNED IF FOUND.

int FSFindEntry(clfs_char *name,int CaseInsens, struct FS_FILE *record,struct FS_FILE *dir)
{
    if(dir==NULL) return FS_ERROR;
    int error;
    struct FS_FILERECORD *fr;
    clfs_64bit PrevDirEntry,DirEntryOffset;
    clfs_32bit nCRC,nCRCfolded;

    // COMPUTE CRC OF NAME TO SEARCH FOR
    nCRC=FSCalcNameCRC(name);
    nCRCfolded=FSCalcFoldedNameCRC(name);

    PrevDirEntry=DirEntryOffset=dir->F.FileSize;

// BUFFERS MUST BE INITIALIZED SOMEWHERE ELSE
//    error=FSInitFileBuffers(dir);
//    if(error!=FS_OK) return error;

    DirEntryOffset-=DIRENTRY_SIZE;
    if(DirEntryOffset<0) return FS_NOTFOUND; // LAST ENTRY ALREADY FOUND
    fr=(struct FS_FILERECORD *)FSBufferMap(dir,DirEntryOffset);
    while(fr) {
        if ( (fr->EntryType<PRIMARY_ENTRY_LIMIT) )  {
            // FOUND PRIMARY ENTRY, FIRST CHECK CRC THEN NAME
            if(fr->EntryType==DE_FILE) {
            // CHECK NAME CRC FIRST

            // DECIDE WHICH CRC TO USE:
                if(CaseInsens || (fr->Attributes&FSATTR_CASEINSENS))
                {
                    // CASE INSENSITIVE COMPARISON
                    if( (fr->FoldedCRC==0) || (nCRCfolded==fr->FoldedCRC))
                    {
                    // CRC MATCH, NOW DO A REAL NAME COMPARISON
                        clfs_32bit CharCache;   // FOR PARTIAL UNICODE CHARACTERS CUT BETWEEN DIRECTORY ENTRIES
                        clfs_char *nameptr;
                        struct FS_NAMERECORD *nr;
                        clfs_64bit InternalOffset;
                        int isamatch;

                        CharCache=0;
                        isamatch=FSPartialNameCompare(name,fr->Name,&CharCache,&nameptr,DIRENTRY_NAME_LENGTH,FS_TRUE);
                        nr=(struct FS_NAMERECORD *)fr;
                        InternalOffset=DirEntryOffset;
                        while(isamatch && (*nameptr!=0)) {
                            ++nr;
                            InternalOffset+=DIRENTRY_SIZE;
                            if(InternalOffset>=PrevDirEntry) break;
                            if((clfs_byte *)nr>=dir->Buffer.Data+dir->Volume->ClusterSize) nr=(struct FS_NAMERECORD *)FSBufferMap(dir,InternalOffset);
                            if(!nr) break;  // THIS SHOULD ONLY HAPPEN IF WE ARE OUT OF MEMORY OR SEVERE DATA CORRUPTION
                            if(nr->EntryType==DE_NAME) isamatch=FSPartialNameCompare(nameptr,nr->Name,&CharCache,&nameptr,DIRENTRY_SIZE-4,FS_TRUE);
                        }

                        if(isamatch && (*nameptr==0)) {
                            // NAME IS A MATCH, FILL STRUCTURE AND RETURN
                            error=FSGetDirEntryData(dir,DirEntryOffset,record);
                            return error;
                           }
                        // NAME COMPARISON MAY HAVE CHANGED THE BUFFER CONTENTS TO ANOTHER CLUSTER, ENSURE THE PROPER CLUSTER IS MAPPED.
                        fr=(struct FS_FILERECORD *)FSBufferMap(dir,DirEntryOffset);
                    }
                }
                else {
                    // CASE SENSITIVE COMPARISON
                    if((fr->NameCRC==0) || (nCRC==fr->NameCRC))
                    {
                        // CRC MATCH, NOW DO A REAL NAME COMPARISON
                        // CRC MATCH, NOW DO A REAL NAME COMPARISON
                            clfs_32bit CharCache;   // FOR PARTIAL UNICODE CHARACTERS CUT BETWEEN DIRECTORY ENTRIES
                            clfs_char *nameptr;
                            struct FS_NAMERECORD *nr;
                            clfs_64bit InternalOffset;
                            int isamatch;

                            CharCache=0;
                            isamatch=FSPartialNameCompare(name,fr->Name,&CharCache,&nameptr,DIRENTRY_NAME_LENGTH,FS_FALSE);
                            nr=(struct FS_NAMERECORD *)fr;
                            InternalOffset=DirEntryOffset;
                            while(isamatch && (*nameptr!=0)) {
                                ++nr;
                                InternalOffset+=DIRENTRY_SIZE;
                                if(InternalOffset>=PrevDirEntry) break;
                                if((clfs_byte *)nr>=dir->Buffer.Data+dir->Volume->ClusterSize) nr=(struct FS_NAMERECORD *)FSBufferMap(dir,InternalOffset);
                                if(!nr) break;  // THIS SHOULD ONLY HAPPEN IF WE ARE OUT OF MEMORY OR SEVERE DATA CORRUPTION
                                if(nr->EntryType==DE_NAME) isamatch=FSPartialNameCompare(nameptr,nr->Name,&CharCache,&nameptr,DIRENTRY_SIZE-4,FS_FALSE);
                            }

                            if(isamatch && (*nameptr==0)) {
                                // NAME IS A MATCH, FILL STRUCTURE AND RETURN
                                error=FSGetDirEntryData(dir,DirEntryOffset,record);
                                return error;
                               }
                            // NAME COMPARISON MAY HAVE CHANGED THE BUFFER CONTENTS TO ANOTHER CLUSTER, ENSURE THE PROPER CLUSTER IS MAPPED.
                            fr=(struct FS_FILERECORD *)FSBufferMap(dir,DirEntryOffset);

                    }

                }

            }

            // SIMPLY CONTINUE...
            PrevDirEntry=DirEntryOffset;
        }
        DirEntryOffset-=DIRENTRY_SIZE;
        if(DirEntryOffset<0) return FS_NOTFOUND; // LAST ENTRY
        --fr;
        if((clfs_byte *)fr<dir->Buffer.Data) fr=(struct FS_FILERECORD *)FSBufferMap(dir,DirEntryOffset);
    }

    // THIS CAN ONLY HAPPEN WITH SEVERE DATA CORRUPTION
    return FS_ERROR;

}


// FILL THE file STRUCTURE WITH ALL DATA CONTAINED IN THE DIRECTORY ENTRY AT Offset

int FSGetDirEntryData(struct FS_FILE *dir,clfs_64bit Offset,struct FS_FILE *file)
{
int error;
struct FS_FILERECORD *fr;
struct FS_RECORDLIST *newrec;
clfs_32bit nentrycount;
// CLEAR THE STRUCTURE;
FSMemset(file,0,sizeof(struct FS_FILE));


fr=(struct FS_FILERECORD *)FSBufferMap(dir,Offset);

//        struct FS_FILERECORD F;

FSMemcpy(&file->F,fr,sizeof(struct FS_FILERECORD));


// LEAVE SOME FIELD TO THE DEFAULT VALUE ZERO
//        clfs_32bit   Mode;
//        clfs_64bit CurrentOffset;
//        clfs_char *Name;
// struct FS_BUFFER Buffer;

file->Volume=dir->Volume;
file->Dir=dir;
file->DirEntryOffset=Offset;
file->DirEntryCount=1;      // WILL INCREMENT AS WE SCAN MORE SECONDARY ENTRIES

Offset+=DIRENTRY_SIZE;
if(Offset>=dir->F.FileSize) fr=NULL;
else {
++fr;
if((clfs_byte *)fr>=dir->Buffer.Data+dir->Volume->ClusterSize) fr=(struct FS_FILERECORD *)FSBufferMap(dir,Offset);
}
error=FS_OK;
nentrycount=0;

while(fr)
{

    if(fr->EntryType<PRIMARY_ENTRY_LIMIT) break;    // END OF FILE RECORD

    newrec=FSMalloc(sizeof(struct FS_RECORDLIST));
    if(!newrec) {  error=FS_NOMEM; break; }
    newrec->Next=NULL;
    FSMemcpy(&newrec->Filerec,fr,DIRENTRY_SIZE);
    if(fr->EntryType==DE_NAME) ++nentrycount;
    if(fr->EntryType==DE_FRAGMENT) FSAppendFragments(file,newrec);
    else FSAppendMetadata(file,newrec);
    file->DirEntryCount++;
    Offset+=DIRENTRY_SIZE;
    if(Offset>=dir->F.FileSize) break;
    ++fr;
    if((clfs_byte *)fr>=dir->Buffer.Data+dir->Volume->ClusterSize) fr=(struct FS_FILERECORD *)FSBufferMap(dir,Offset);

}


// CREATE A CLUSTER CHAIN IF NEEDED
if(!file->Chain) {
    if(file->F.NumClusters) {
        // THE FILE HAS STORAGE BUT NO FRAGMENTS. CREATE A CHAIN
        file->Chain=(struct FS_RECORDLIST *)FSMalloc(sizeof(struct FS_RECORDLIST));  // INITIALIZE 1-CLUSTER STORAGE
        if(!file->Chain) error=FS_NOMEM;
        else {
        FSMemset(file->Chain,0,sizeof(struct FS_RECORDLIST));
        file->Chain->Fragrec.EntryType=DE_FRAGMENT;
        file->Chain->Fragrec.Fragments[0].StartCluster=file->F.StartCluster;
        file->Chain->Fragrec.Fragments[0].NumClusters=file->F.NumClusters;
        }
    }
}



// FINALLY, EXTRACT NAME FROM METADATA
file->Name=FSMalloc(nentrycount*NAMERECORD_LENGTH+DIRENTRY_NAME_LENGTH+1);
if(!file->Name) error=FS_NOMEM;
else {
    clfs_char *nameptr=file->Name;
    FSMemcpy(nameptr,file->F.Name,DIRENTRY_NAME_LENGTH);
    nameptr+=DIRENTRY_NAME_LENGTH;

    newrec=file->Metadata;
    while(newrec)
    {
        if(newrec->Namerec.EntryType==DE_NAME) {
        FSMemcpy(nameptr,newrec->Namerec.Name,NAMERECORD_LENGTH);
        nameptr+=NAMERECORD_LENGTH;
        }
        newrec=newrec->Next;
    }
    *nameptr=0; // ADD NULL TERMINATOR
}


if(error!=FS_OK) {
// CLEANUP
if(file->Chain) {
    struct FS_RECORDLIST *ptr;
    do { ptr=file->Chain; file->Chain=file->Chain->Next; FSFree(ptr); } while(file->Chain);
}
if(file->Metadata) {
    struct FS_RECORDLIST *ptr;
    do { ptr=file->Metadata; file->Metadata=file->Metadata->Next; FSFree(ptr); } while(file->Metadata);
}
if(file->Name) { FSFree(file->Name); file->Name=NULL; }
}



return error;

}

int FSSeek(struct FS_FILE *file,clfs_64bit Offset,clfs_32bit position)
{
clfs_64bit from=0;
if(position==SEEK_END) from=file->F.FileSize;
if(position==SEEK_CUR) from=file->CurrentOffset;

from+=Offset;

if(from<0) from=0;
if(from>file->F.FileSize) from=file->F.FileSize;
file->CurrentOffset=from;
return FS_OK;

}

void FSMemset(void *dest,clfs_32bit value,clfs_32bit nbytes)
{
    // WRITTEN IN C, OPTIMIZED FOR ARM


    clfs_32bit val32= (value&0xff) | ((value&0xff)<<8)
                                | ((value&0xff)<<16)
                                | ((value&0xff)<<24);
    clfs_64bit val64= val32 | (((clfs_64bit)val32)<<32);

    // FIRST MISALIGNED BYTES
    clfs_32bit destaddr=((clfs_32bit)dest)&0x7;
    clfs_byte *byteptr=(clfs_byte *)dest;
    while( (destaddr&7) &&(nbytes)) {
    *byteptr=value&0xff;
    ++destaddr;
    ++byteptr;
    --nbytes;
    }

    // DO ALIGNED 64-BIT TRANSFERS
    clfs_64bit *ptr64=(clfs_64bit *)byteptr;
    // 64-BYTE BLOCK TRANSFERS
    while(nbytes>=64) {
        *ptr64=val64;
        ptr64[1]=val64;
        ptr64[2]=val64;
        ptr64[3]=val64;
        ptr64[4]=val64;
        ptr64[5]=val64;
        ptr64[6]=val64;
        ptr64[7]=val64;
        ptr64+=8;
        nbytes-=64;
    }
    // 8-BYTE BLOCK TRANSFERS
    while(nbytes>=8) {
        *ptr64=val64;
        ++ptr64;
        nbytes-=8;
    }
    // FINAL MISALIGNED BYTES
    byteptr=(clfs_byte *)ptr64;
    while(nbytes) {
    *byteptr=value&0xff;
    ++byteptr;
    --nbytes;
    }

}

void FSMemcpy(void *dest,void *src,clfs_32bit nbytes)
{
    // OPTIMIZED FOR ARM WITH WORD-ALIGNED READ/WRITES


   clfs_32bit destaddr=((clfs_32bit)dest)&0x3;
   clfs_32bit srcaddr=((clfs_32bit)src)&0x3;

   if( (!destaddr) && (!srcaddr) ) {
       // WORD-ALIGNED COPY
       // DO ALIGNED 64-BIT TRANSFERS
       clfs_64bit *dest64=(clfs_64bit *)dest;
       clfs_64bit *src64=(clfs_64bit *)src;
       // 64-BYTE BLOCK TRANSFERS
       while(nbytes>=64) {
           *dest64=*src64;
           dest64[1]=src64[1];
           dest64[2]=src64[2];
           dest64[3]=src64[3];
           dest64[4]=src64[4];
           dest64[5]=src64[5];
           dest64[6]=src64[6];
           dest64[7]=src64[7];
           dest64+=8;
           src64+=8;
           nbytes-=64;
       }
       // 8-BYTE BLOCK TRANSFERS
       while(nbytes>=8) {
           *dest64=*src64;
           ++dest64;
           ++src64;
           nbytes-=8;
       }
       // FALL-OVER TO BYTE ALIGNED COPY FOR LAST FEW MISALIGNED BYTES
       dest=(void *)dest64;
       src=(void *)src64;
   }

   // BYTE-ALIGNED COPY (SLOWER)
       clfs_byte *destptr=(clfs_byte *)dest,*srcptr=(clfs_byte *)src;

       while(nbytes) {
           *destptr++=*srcptr++;
           --nbytes;
       }

}


// LOW-LEVEL VERSION OF OPENDIR, ONLY FILLS FS_FILE STRUCTURE
// DOES NOT PUT DIRECTORY IN OPEN FILE LIST (CALLING FUNCTION IS RESPONSIBLE FOR THAT)
// DESIGNED TO BE ABLE TO OPEN THE ROOT DIRECTORY WHILE MOUNTING
// DOES **NOT** FILL THE FOLLOWING MEMBERS (CALLING FUNCTION IS RESPONSIBLE FOR THAT)
// Dir = Pointer to the parent directory
// Name = Name of the directory
// DirEntryOffset = Offset within the parent directory to update the entry
// DirEntryCount = Number of records in parent directory

int FSOpenDirLL(struct FS_FILE *Dir,struct FS_VOLUME *fs,clfs_64bit FirstCluster)
{
    // FIRST STAGE: POPULATE TEMPORARY FS_FILE STRUCTURE WITH 1-CLUSTER LONG DIRECTORY

    struct FS_RECORDLIST *temp;

    int error;

    //FSMemset(Dir,0,sizeof(struct FS_FILE));    // ZERO-OUT ENTIRE STRUCTURE

    Dir->Mode=FSMODE_MODIFY|FSMODE_WRITE;    // DEFAULT DIRECTORY ACCESS IS MODIFY
    Dir->F.Attributes=FSATTR_DIR;       // LATER WILL GET OTHER ATTRIBUTES
    Dir->F.FileSize=fs->ClusterSize; // START WITH ONE-CLUSTER
    Dir->F.StartCluster=FirstCluster;
    Dir->F.NumClusters=1;         // TEMPORARY 1-CLUSTER CHAIN
//  Dir->Name=NULL;             // LATER WILL STORE NAME FROM PARENT DIRECTORY
    Dir->Volume=fs;
    Dir->Metadata=NULL;
//  Dir->CurrentOffset=0;
//  Dir->DirEntryOffset=0;      // LATER WILL STORE THIS INFORMATION
//  Dir->DirEntryCount=0;
//  Dir->Dir=NULL;              // LEAVE FOR LATER
    Dir->Buffer.Data=NULL;

    error=FSInitFileBuffers(Dir);   // INITIALIZE READ/WRITE BUFFER
    if(error!=FS_OK) return error;

    Dir->Chain=(struct FS_RECORDLIST *)FSMalloc(sizeof(struct FS_RECORDLIST));  // INITIALIZE 1-CLUSTER STORAGE
    if(!Dir->Chain) return FS_NOMEM;
    FSMemset(Dir->Chain,0,sizeof(struct FS_RECORDLIST));
    Dir->Chain->Fragrec.EntryType=DE_FRAGMENT;
    Dir->Chain->Fragrec.Fragments[0].StartCluster=FirstCluster;
    Dir->Chain->Fragrec.Fragments[0].NumClusters=1;


    // READ FIRST CLUSTER DIRECTLY INTO READ BUFFER
    int entryfound=0,firstfrag=0;
    struct FS_FILERECORD *ptr,*endofbuffer;
    clfs_64bit Off;

    Off=0;
    endofbuffer=(struct FS_FILERECORD *)(Dir->Buffer.Data+fs->ClusterSize);

    // START READING CLUSTERS ONE BY ONE
    while(FSBufferFill(Dir,Off)==FS_OK) {
        // NOW SCAN FIRST CLUSTER FOR THE '.' ENTRY
        ptr=(struct FS_FILERECORD *)Dir->Buffer.Data;
        if(!entryfound) {
        while(ptr!=endofbuffer) {
            if(ptr->EntryType!=DE_FILE) { ++ptr; continue; }
            if( (ptr->Attributes&FSATTR_DIR) && (ptr->Name[0]=='.') && (ptr->Name[1]==0) && (ptr->StartCluster==FirstCluster)) {
                entryfound=1;
                // GET MORE INFORMATION
                FSMemcpy(&(Dir->F),ptr,sizeof(struct FS_FILERECORD));

                Dir->Chain->Fragrec.Fragments[0].NumClusters=ptr->NumClusters;  // INCREASE NUMBER OF CLUSTERS ON DIRECTORY
                Dir->DirEntryCount2=1;
                Dir->DirEntryOffset2=Off+(clfs_64bit)((clfs_byte *)ptr-Dir->Buffer.Data);
                ++ptr;
                break;
            }
            ++ptr;
            continue;
        }
        }
        if(entryfound) {
            while(ptr!=endofbuffer) {
               // NOW CONTINUE SCANNING FOR ALL OTHER ENTRIES
                if(ptr->EntryType<PRIMARY_ENTRY_LIMIT) {
                    // ANOTHER PRIMARY ENTRY MEANS THE END OF THIS ENTRY
                    // JUST RETURN
                    return FS_OK;
                }

                temp=FSMalloc(sizeof(struct FS_RECORDLIST));
                if(!temp) {
                    while(Dir->Chain) { temp=Dir->Chain; Dir->Chain=Dir->Chain->Next; FSFree(temp); }
                    while(Dir->Metadata) { temp=Dir->Metadata; Dir->Metadata=Dir->Metadata->Next; FSFree(temp); }
                    FSFree(Dir->Buffer.Data);
                    Dir->Buffer.Data=NULL;
                    Dir->Buffer.Used=0;
                    return FS_NOMEM;
                }
                // CREATE NEW RECORDLIST BLOCK
                FSMemcpy(&(temp->Fragrec),ptr,sizeof(struct FS_FRAGRECORD));
                temp->Next=NULL;

                ++Dir->DirEntryCount2;

                if(ptr->EntryType==DE_FRAGMENT) {
                    if(firstfrag) {  // THIS IS NOT THE FIRST FRAGMENT
                    FSAppendFragments(Dir,temp);
                    }
                    else {  // THIS IS THE FIRST FRAGMENT RECORD, SIMPLY OVERWRITE THE TEMPORARY ONE
                        FSMemcpy(&(Dir->Chain->Fragrec),ptr,sizeof(struct FS_FRAGRECORD));
                        FSFree(temp);
                        firstfrag=1;
                    }

                }
                else {
                    // APPEND AS METADATA
                    FSAppendMetadata(Dir,temp);
                }
                ++ptr;
            }

        }
        else break;

     }

    // END OF THE CLUSTER CHAIN WAS REACHED
    if(entryfound) {
        // IT'S OK IN THE (RARE) CASE WHERE THE '.' ENTRY WAS THE LAST IN THE ONLY CLUSTER
        return FS_OK;
    }

    while(Dir->Chain) { temp=Dir->Chain; Dir->Chain=Dir->Chain->Next; FSFree(temp); }
    while(Dir->Metadata) { temp=Dir->Metadata; Dir->Metadata=Dir->Metadata->Next; FSFree(temp); }
    FSFree(Dir->Buffer.Data);
    Dir->Buffer.Data=NULL;
    Dir->Buffer.Used=0;
    return FS_INVALIDDATA;

}


// HELPER FUNCTION RETURNS FREE CLUSTER COUNT AT LEFT, RIGHT AND MIDDLE OF A 32-BIT WORD
inline void FSAnalyzeUsage(clfs_32bit word,clfs_32bit *left,clfs_32bit *mid,clfs_32bit *midpos,clfs_32bit *right)
{
    int f,k;
    clfs_32bit cnt[32];
    clfs_32bit pos[32];

    FSMemset(cnt,0,sizeof(cnt));
    FSMemset(pos,0,sizeof(pos));

    for(f=0,k=0;f<8;++f)
    {

        switch(word&0xf)
        {
        case 0:
            cnt[k]+=4;
            break;
        case 1:
            ++k;
            cnt[k]+=3;
            pos[k]=f*4+1;
            break;
        case 2:
            cnt[k]++;
            ++k;
            cnt[k]+=2;
            pos[k]=f*4+2;
            break;
        case 3:
            ++k;
            cnt[k]+=2;
            pos[k]=f*4+2;

            break;
        case 4:
            cnt[k]+=2;
            ++k;
            cnt[k]+=1;
            pos[k]=f*4+3;

            break;
        case 5:
            ++k;
            cnt[k]++;
            pos[k]=f*4+1;

            ++k;
            cnt[k]++;
            pos[k]=f*4+3;

            break;
        case 6:
            cnt[k]++;
            ++k;
            cnt[k]++;
            pos[k]=f*4+2;
            ++k;
            cnt[k]++;
            pos[k]=f*4+3;

            break;
        case 7:
            ++k;
            cnt[k]++;
            pos[k]=f*4+3;
            break;
        case 8:
            cnt[k]+=3;
            ++k;
            pos[k]=f*4+4;
            break;
        case 9:
            ++k;
            cnt[k]+=2;
            pos[k]=f*4+1;
            ++k;
            pos[k]=f*4+4;
            break;
        case 10:
            cnt[k]++;
            ++k;
            cnt[k]++;
            pos[k]=f*4+2;
            ++k;
            pos[k]=f*4+4;
            break;
        case 11:
            ++k;
            cnt[k]++;
            pos[k]=f*4+2;
            ++k;
            pos[k]=f*4+4;
            break;
        case 12:
            cnt[k]+=2;
            ++k;
            pos[k]=f*4+4;
            break;
        case 13:
            ++k;
            cnt[k]++;
            pos[k]=f*4+1;
            ++k;
            pos[k]=f*4+4;
            break;
        case 14:
            cnt[k]++;
            ++k;
            pos[k]=f*4+4;
            break;
        case 15:
            ++k;
            pos[k]=f*4+4;
            break;
        }



        word>>=4;
    }

    *left=cnt[0];
    *right=cnt[k];
    *mid=0;
    for(f=1;f<k;++f)
    {
        if(*mid<cnt[k]) { *mid=cnt[k]; *midpos=pos[k]; }
    }

}

// SCAN FREE SPACE BITMAP TO FIND SUITABLE AREA FOR FAST CLUSTER ALLOCATION
// RETURN -1 ON ERROR

clfs_64bit FSNextFreeCluster(struct FS_VOLUME *fs)
{

    clfs_32bit mask,left,mid,right,midpos;
    clfs_32bit *ptr,*endofbuffer;
    clfs_64bit Offset,Offsetstart,cluster;
    clfs_64bit zonestart,maxzonestart,maxcount,count,pos;
    int error;


// INITIALIZATION IS GUARANTEED BY FSAllocClusters
//    error=FSInitFileBuffers(fs->FreeClusters);
//    if(error!=FS_OK) return error;


    endofbuffer=(clfs_32bit *)(fs->FreeClusters->Buffer.Data+fs->ClusterSize);

    cluster=fs->NextFreeClusterHint;

    mask=1<<(cluster&31);
    Offset=(cluster>>3)&(~(fs->ClusterSize-1));
    ptr=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
    ptr+=((cluster>>3)&(fs->ClusterSize-1))>>2;        // FIND PROPER WORD WITHIN BUFFER

    // READ CLUSTER

    error=FSBufferFill(fs->FreeClusters,Offset);
    if(error!=FS_OK) return -1;

    // HAVE PROPER CLUSTER IN BUFFER

    if( (!(*ptr)) & mask) {
        return fs->NextFreeClusterHint;
    }

    // GET NEXT LARGE ZONE OF FSFree CLUSTERS
    // START SCANNING CURRENTLY LOADED CLUSTER
    Offsetstart=Offset;
    maxzonestart=zonestart=Offset*8;
    maxcount=count=0;

    do {
    ptr=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
    pos=Offset*8;

    // SCAN ENTIRE CLUSTER LOOKING FOR LARGE AREA TO STORE FILES
    do {

        if(!(*ptr)) {
            // CLUSTERS ARE ALL FSFree
            count+=32;
        }
        else {
            FSAnalyzeUsage(*ptr,&left,&mid,&midpos,&right);
            count+=left;
            if(count>maxcount) {
                maxzonestart=zonestart;
                maxcount=count;
            }
            if(mid>maxcount) {
                maxzonestart=pos+midpos;
                maxcount=mid;
            }
            zonestart=pos+32-right;
            count=right;
            }
        ++ptr;
        pos+=32;
    } while(ptr!=endofbuffer);


                // ADVANCE TO NEXT CLUSTER
                Offset+=fs->ClusterSize;
                if(Offset>=fs->FreeClusters->F.FileSize) {
                    Offset=0;        // CIRCLE BACK TO START OF VOLUME AND KEEP LOOKING
                    // RESET COUNT
                    if(count>maxcount) {
                        maxzonestart=zonestart;
                        maxcount=count;
                    }
                    zonestart=0;
                    count=0;
                }

                if(Offset==Offsetstart) break;  // FULL CIRCLE, WHOLE DISK WAS SCANNED, LEAVE EARLY TO AVOID UNNECESSARY READ INTO BUFFER

                error=FSBufferFill(fs->FreeClusters,Offset);
                if(error!=FS_OK) return -1;

                    ptr=(clfs_32bit *)fs->FreeClusters->Buffer.Data;    // POINT TO START OF NEW BUFFERR
    } while(Offset!=Offsetstart);


    // DONE, WHOLE DISK WAS SCANNED AND LARGEST AREA WAS FOUND

    fs->NextFreeClusterHint=maxzonestart;
    return maxzonestart;

}




// RETURN TYPE OF PATH
// 1 == Name is empty
// 2 == Name include path
// 4 == Path is absolute
// 8 == File ends in slash
// BITS 8-31 = DEPTH OF PATH (0=CURRENT DIR OR ROOT)
// NEGATIVE RESULT ==> ERROR CODE (NORMALLY FS_BADNAME)

// FILENAME RESTRICTIONS ON CleanFS:

// 1) NAME MUST BE A CORRECTLY FORMATTED UTF-8 STRING
// 2) NAME MUST NOT INCLUDE THE FORWARD SLASH '/'       (NOT PERMITTED IN LINUX/WINDOWS OR OSX)

// OPTIONAL API IN CleanFS CAN HELP REMOVE CHARACTERS IN THE RANGES:
/*          0000 - 001F     COMMON ASCII CONTROL CHARS (INCLUDING NEWLINE)
            007F            ANOTHER ASCII CONTROL CHAR
            0080 - 009F     ISO 6429 CONTROL CODES (C1 CONTROL CODES)
            2028            LINE SEPARATOR
            2029            PARAGRAPH SEPARATOR
            FEFF            BYTE ORDER MARKER
            FFF0 - FFFF     VARIOUS SPECIAL CHARACTERS, INCLUDING REVERSE BYTE ORDER MARKER
            FDD0 - FDEF     UNICODE NON-CHARACTERS
          nnFEFF            SPECIAL CHARACTER IN ALL PLANES
          nnFFFF            SPECIAL CHARACTER IN ALL PLANES
*/



int FSGetNameType(clfs_char *name)
{
int flags=0;

int depth=0;
#ifdef FSCONFIG_WINDOWS_COMPAT
const clfs_char illegal[]=":*<>\"?|";
const clfs_char separators[]="\\/";
#else
const clfs_char illegal[]=":";
const clfs_char separators[]="/";
#endif

clfs_char *ptr,*partial;

partial=name;
if(!__fsisvalidUTF8(partial)) return FS_BADNAME;           // CHECK FOR VALID UTF-8 SEQUENCES
if(__fshasinvalidcharsUTF8(partial)) return FS_BADNAME;     // CHECK FOR ILLEGAL CONTROL CHARACTERS
if(__fsfindchar(partial,NULL,illegal)) return FS_BADNAME;	// NAME CANNOT CONTAIN ILLEGAL CHARACTERS

partial=name;
while((ptr=__fsfindchar(partial,NULL,separators))) {

// PATH SEPARATOR FOUND
flags|=2;
if( ptr==name ) flags|=4;		// PATH IS ABSOLUTE
else {
if(ptr==partial) return FS_BADNAME;	// TWO SLASHES TOGETHER
++depth;
}
if(ptr[1]==0) flags|=8;		// NAME ENDS WITH SLASH

partial=ptr+1;
}

if(*partial==0) flags|=1;		// NAME IS EMPTY
return flags | (depth<<8);

}




// PARSE NAME, FIND FILE AND FILL THE ENTRY

int FSFindFile(clfs_char *name,struct FS_FILE *entry, struct FS_VOLUME *fs,clfs_32bit dirsvalid)
{
    int ntype,f;
    int error;
    clfs_char *ptr,*ptrend,*temp;
    struct FS_FILE *startdir,*dir,*newdir;
    struct FS_RECORDLIST *fr;
    struct FS_FILELIST *lst;
#ifdef FSCONFIG_WINDOWS_COMPAT
    const clfs_char separators[]="\\/";
#else
    const clfs_char separators[]="/";
#endif



    ntype=FSGetNameType(name);
    //printf("name %s type=%02X\n",name,ntype);

    if(ntype==FS_BADNAME) return FS_BADNAME;

    if(!dirsvalid) {
        if((ntype&FSNAME_ENDSLASH) || (ntype&FSNAME_EMPTY)) return FS_BADNAME;
    }

    if(!fs) return FS_NODISK;

    ptr=name;




    // OBTAIN STARTING DIRECTORY


    if(ntype&FSNAME_ISABSOLUTE) {
        startdir=&fs->RootDir;
        ++ptr;		// SKIP THE INITIAL SLASH
    }
    else startdir=fs->CurrentDir;

    //printf("partial=%s\n",ptr);
    dir=startdir;

    if(ntype&FSNAME_HASPATH) {
        // CRAWL DIRECTORIES UNTIL WE REACH THE FINAL NAME
        //printf("start crawl\n");
        // GET STRING SPACE AT LEAST AS LONG AS THE FULL PATH GIVEN TO US
        temp=(clfs_char *)FSMalloc(__fsstrlen(name)+1);
        if(!temp) return FS_NOMEM;


        while((ptrend=__fsfindchar(ptr,NULL,separators))) {
            // EXTRACT DIRECTORY NAME
            FSMemcpy(temp,ptr,ptrend-ptr);
            temp[ptrend-ptr]=0;
            newdir=NULL;
            //printf("CD %s\n",temp);
            if( (temp[0]=='.')&&(temp[1]==0)) {		// IGNORE REFERENCES TO CURRENT DIR
                ptr=ptrend+1;
                continue;
            }
            if((temp[0]=='.')&&(temp[1]=='.')&&(temp[2]==0)) {
                newdir=dir;
                dir=dir->Dir;				// GET PARENT DIRECTORY WITHOUT SCANNING
                FSFreeFile(newdir);
                if(dir==NULL) { FSFree(temp); return FS_NOTFOUND; }		// GONE BEYOND ROOTDIR;
                ptr=ptrend+1;
                continue;
            }

            // CHECK FOR ALREADY OPEN DIRECTORIES IN THIS VOLUME

            lst=fs->OpenFiles;
            while(lst)
            {
                // CHECK IF THIS FILE IS OR CONTAINS THE PATH WE NEED
                for(f=0;f<lst->UsedCount;++f) {
                    newdir=lst->File[f];
                while(newdir) {
                    if(newdir->Dir==dir) {
                        // CHECK FOR SAME PARENT DIR AND SAME NAME
                        if((newdir->Name!=NULL) && FSNameCompare(temp,newdir->Name,(newdir->F.Attributes&FSATTR_CASEINSENS))) { break; }
                    }
                    newdir=newdir->Dir;
                }
                if(newdir) break;
                }
                lst=lst->Next;
            }
            if(newdir) {
                ptr=ptrend+1;
                continue;
            }
            // CHECK FOR CURRENTDIR
            if(!newdir) {
                newdir=fs->CurrentDir;
                while(newdir) {
                    if(newdir->Dir==dir) {
                        // CHECK FOR SAME PARENT DIR AND SAME NAME
                        if((newdir->Name!=NULL) && FSNameCompare(temp,newdir->Name,(newdir->F.Attributes&FSATTR_CASEINSENS))) break;
                    }
                    newdir=newdir->Dir;
                }
            }

            if(!newdir) {
                // IF IT'S NOT OPEN YET, SCAN THE DIRECTORY
                newdir=(struct FS_FILE *)FSMalloc(sizeof(struct FS_FILE));
                if(!newdir) {
                    while(dir!=NULL) dir=FSFreeFile(dir);
                    FSFree(temp);
                    return FS_NOMEM;
                }

                // CLEAN ENTRY
                FSMemset((void *)newdir,0,sizeof(struct FS_FILE));

                error=FSFindEntry(temp,FS_FALSE,newdir,dir);
                if(error!=FS_OK) {
                    // CLEANUP PROCEDURE
                    while(dir!=NULL) dir=FSFreeFile(dir);
                    FSFree(temp);
                    FSFree(newdir);
                    if(error==FS_EOF) return FS_NOTFOUND;
                    return error;
                }




            if(!(newdir->F.Attributes&FSATTR_DIR)) {	// FILE FOUND BUT IT'S NOT A DIRECTORY
                while(newdir!=NULL) newdir=FSFreeFile(newdir);
                FSFree(temp);
                return FS_NOTFOUND;
            }

            // OPEN DIRECTORY ENTRY

            // RELEASE ORIGINAL CLUSTER CHAIN MEMORY


            // FREE ENTIRE CLUSTER CHAIN AND METADATA, BUT KEEP THE NAME
            while(newdir->Chain) { fr=newdir->Chain->Next; FSFree(newdir->Chain); newdir->Chain=fr; }
            while(newdir->Metadata) { fr=newdir->Metadata->Next; FSFree(newdir->Metadata); newdir->Metadata=fr; }

            error=FSOpenDirLL(newdir,fs,newdir->F.StartCluster);

            if(error!=FS_OK) {
                while(newdir!=NULL) newdir=FSFreeFile(newdir);
                FSFree(temp);
                return error;
            }

            // FSOpenDirLL DOES **NOT** FILL THE FOLLOWING MEMBERS (CALLING FUNCTION IS RESPONSIBLE FOR THAT)
            // Dir = Pointer to the parent directory
            // Name = Name of the directory

            // THESE ENTRIES WERE PREVIOUSLY FILLED BY FSFindEntry

            }
            // DO NEXT DIR
            dir=newdir;
            ptr=ptrend+1;

        }

        FSFree(temp);

    }


    // HERE DIR IS THE DIRECTORY OF THE FILE
    // ptr POINTS TO THE FILE NAME

    //printf("final name=%s\n",ptr);

    if((ptr[0]=='.')&&(ptr[1]==0)) {		// NAME IS CURRENT DIR
        //printf("dot\n");
        ntype|=FSNAME_EMPTY|FSNAME_HASPATH;
    }
    else if((ptr[0]=='.')&&(ptr[1]=='.')&&(ptr[2]==0)) {
        //printf("dotdot\n");
        newdir=dir;
        ntype|=FSNAME_EMPTY|FSNAME_HASPATH;
        dir=dir->Dir;				// GET PARENT DIRECTORY WITHOUT SCANNING
        FSFreeFile(newdir);
        if(dir==NULL) { return FS_NOTFOUND; }		// GONE BEYOND ROOTDIR;
    }


    if(ntype&FSNAME_EMPTY) {		// NAME IS EMPTY

        if(!dirsvalid) {
            while(dir!=NULL) dir=FSFreeFile(dir);

            return FS_NOTFOUND;
        }
        //printf("empty name w/dirsvalid\n");
        if(!(ntype&FSNAME_HASPATH)) {
            // EMPTY STRING PASSED AS ARGUMENT
            //printf("null name\n");
            while(dir!=NULL) dir=FSFreeFile(dir);

            return FS_NOTFOUND;
        }
        // RETURN DIRECTORY FOUND
        if(FSFileIsReferenced(dir,fs)) {
            // RETURN EXISTING DIRECTORY
            //printf("existing dir returned\n");
            FSMemset((void *)entry,0,sizeof(struct FS_FILE));
            entry->Dir=dir;

            return FS_OPENDIR;
        }
        // RETURN NEWLY CREATED DIRECTORY
        //printf("new dir returned\n");
        FSMemcpy(entry,dir,sizeof(struct FS_FILE));
        FSFree(dir);

        return FS_OK;
    }



    // CHECK TO SEE IF THIS FILE DIRECTORY IS ALREADY OPEN OR REFERENCED
    lst=fs->OpenFiles;
    newdir=NULL;
    while(lst)
    {
        for(f=0;f<lst->UsedCount;++f)
        {
        // CHECK IF THIS FILE IS OR CONTAINS THE PATH WE NEED
            newdir=lst->File[f];
        while(newdir) {
           if(newdir->Dir==dir) {
                // CHECK FOR SAME PARENT DIR AND SAME NAME
                if((newdir->Name!=NULL) && FSNameCompare(ptr,newdir->Name,newdir->F.Attributes&FSATTR_CASEINSENS)) { while(lst->Next) lst=lst->Next; f=lst->UsedCount; break; }
            }
            newdir=newdir->Dir;
        }
        }
        lst=lst->Next;
    }

    if(newdir) {
        // FILE IS OPEN, RETURN CURRENT ENTRY
        FSMemset((void *)entry,0,sizeof(struct FS_FILE));
        // RETURN NOT FOUND IF FILE IS A DIRECTORY
        if(!dirsvalid && (newdir->F.Attributes&FSATTR_DIR)) { return FS_NOTFOUND; }

        // RETURN EXISTING OPEN FILE RECORD
        entry->Dir=newdir;

        return FS_OPENDIR;

    }

    // CHECK IF THIS FILE IS THE CURRENT DIRECTORY (ONLY IF dirsvalid IS TRUE)

    if(dirsvalid) {
        newdir=fs->CurrentDir;
        while(newdir) {

        if(newdir->Dir==dir) {
            // CHECK FOR SAME PARENT DIR AND SAME NAME
            if((newdir->Name!=NULL) && FSNameCompare(ptr,newdir->Name,newdir->F.Attributes&FSATTR_CASEINSENS)) {
                // FILE IS OPEN, RETURN CURRENT ENTRY
                FSMemset((void *)entry,0,sizeof(struct FS_FILE));

                // RETURN EXISTING OPEN FILE RECORD
                entry->Dir=newdir;
                return FS_OPENFILE;
            }


        }
        newdir=newdir->Dir;
        }


    }


    // FILE WASN'T OPEN, TRY TO FIND ON DISK


    error=FSFindEntry(ptr,FS_FALSE,entry,dir);

    //printf("findentry returned %d\n",error);
    if(error!=FS_OK) {
        // CLEANUP PROCEDURE
        while(dir!=NULL) dir=FSFreeFile(dir);

        if(error==FS_EOF) return FS_NOTFOUND;
        return error;
    }

    if(!dirsvalid && (entry->F.Attributes&FSATTR_DIR)) {
        // FREE ALSO ALL METADATA AND FRAGMENTS
        FSFreeEntryData(entry);
        FSMemset((void *)entry,0,sizeof(struct FS_FILE));
        while(dir!=NULL) dir=FSFreeFile(dir);

        return FS_NOTFOUND;
    }


    return FS_OK;

}


// CLEANUP FS_FILE STRUCTURES INTERNAL DATA
void FSFreeEntryData(struct FS_FILE *file)
{
struct FS_RECORDLIST *fr;

// FREE ENTIRE CLUSTER CHAIN
while(file->Chain) { fr=file->Chain->Next; FSFree(file->Chain); file->Chain=fr; }
while(file->Metadata) { fr=file->Metadata->Next; FSFree(file->Metadata); file->Metadata=fr; }


// FREE READ/WRITE BUFFERS
if(file->Buffer.Data) { FSFree(file->Buffer.Data); file->Buffer.Data=NULL; file->Buffer.Used=0; }

// FREE NAME
if(file->Name) { FSFree(file->Name); file->Name=NULL; }
}

// CLEANUP FS_FILE STRUCTURES COMPLETE
void FSFreeEntry(struct FS_FILE *file)
{
    FSFreeEntryData(file);
    // FREE FS_FILE
    FSFree(file);
}


// CLEANUP FS_FILE STRUCTURES
struct FS_FILE *FSFreeFile(struct FS_FILE *file)
{
struct FS_FILE *orig=file->Dir;
struct FS_VOLUME *fs;
int error;

if(file==NULL) {
//printf("Trying to Free() NULL\n");
return NULL;
}


// CHECK FOR REFERENCES BEFORE DELETING
fs=file->Volume;

if(!fs) return NULL;

if(FSFileIsReferenced(file,fs)) return NULL;

// NOT REFERENCED
// IT'S SAFE TO DELETE THIS FILE OBJECT


if(file->Mode&FSMODE_DIRTYDIR) {
    error=FSUpdateDirEntry(file,NULL);
    if(error!=FS_OK) return NULL;
}

// FLUSH BUFFERS IF FILE WAS USED
error=FSBufferFlush(file);
if(error!=FS_OK) return NULL;   // CANNOT FREE IF CLOSING THE FILE FAILED


FSFreeEntry(file);
return orig;
}


// CHECK IF FILE IS IN THE OPEN FILE LIST, RETURNS TRUE/FALSE

int FSFileIsOpen(struct FS_FILE *file,struct FS_VOLUME *fs)
{
struct FS_FILE *ff;
struct FS_FILELIST *lst;
int f;


lst=fs->OpenFiles;
while(lst) {
for(f=0;f<lst->UsedCount;++f)
{
ff=fs->OpenFiles->File[f];
if(ff==file) return FS_TRUE;
ff=ff->Dir;
}
lst=lst->Next;
}

return FS_FALSE;

}


// CHECK IF FILE IS REFERENCED, RETURNS TRUE/FALSE

int FSFileIsReferenced(struct FS_FILE *file,struct FS_VOLUME *fs)
{
struct FS_FILE *ff;
struct FS_FILELIST *lst;
int f,refcount=0;

if(file==&fs->RootDir) ++refcount;

ff=fs->CurrentDir;
while(ff!=NULL) {
if(ff==file) ++refcount;
ff=ff->Dir;
}

lst=fs->OpenFiles;
while(lst) {
for(f=0;f<lst->UsedCount;++f)
{
ff=fs->OpenFiles->File[f];
while(ff!=NULL) {
if(ff==file) ++refcount;
ff=ff->Dir;
}
}
lst=lst->Next;
}

return refcount;

}


// FINAL USER-API

int FSOpen(clfs_char *name, clfs_32bit mode, struct FS_FILE **fileptr, struct FS_VOLUME *fs)
{
struct FS_FILE *entry;
int error;

*fileptr=NULL;

if(fs==NULL) return FS_NODISK;

entry=(struct FS_FILE *)FSMalloc(sizeof(struct FS_FILE));
if(entry==NULL) return FS_NOMEM;

// CLEAN ENTRY
FSMemset((void *)entry,0,sizeof(struct FS_FILE));


if( mode& (FSMODE_APPEND|FSMODE_MODIFY) ) mode|=FSMODE_WRITE; // APPEND/MODIFY IMPLY WRITE

if( (mode&FSMODE_WRITE) && (fs->Flags&FSATTR_IMMUTABLE)) return FS_CANTWRITE;

error=FSFindFile(name,entry,fs,FS_FALSE);
if(error!=FS_OK) {

if(			(error==FS_NOTFOUND) &&
                        ( mode &FSMODE_WRITE ) &&
                        !(mode&FSMODE_NOCREATE)
                        )
{
// CREATE FILE
//printf("try to create\n");
error=FSCreate(name,0,fileptr,fs);
//printf("create=%d\n",error);
if(error==FS_OK) {
(*fileptr)->Mode=mode;
}
}

FSFree(entry); return error;


}


entry->Mode=mode;

// TODO: CHECK USER PERMISSION SETTINGS HERE
if( (entry->F.Attributes&FSATTR_IMMUTABLE) && (mode&FSMODE_WRITE)) {	// CANT OPEN READ ONLY FILES FOR WRITE
// CLEANUP PROCEDURE
while(entry!=NULL) entry=FSFreeFile(entry);
return FS_CANTWRITE;
}

if((mode&(FSMODE_WRITE | FSMODE_APPEND | FSMODE_MODIFY))==FSMODE_WRITE)
{
    entry->F.FileSize=0; // TRUNCATE
    clfs_64bit time=FSTimeDateNow();
    entry->F.ModifiedDate=time>>16;	// UPDATE LAST WRITE DATE IF OPENED FOR WRITING
    entry->F.ExtendedDates=(entry->F.ExtendedDates&0xffff0000)|(time&0xffff);

}

if(mode&FSMODE_APPEND) entry->CurrentOffset=entry->F.FileSize;



error=FSAddToOpenFileList(entry);
if(error!=FS_OK) {
    // CLEANUP PROCEDURE
    while(entry!=NULL) entry=FSFreeFile(entry);
    return error;
}
*fileptr=entry;
return FS_OK;

}

int FSAddToOpenFileList(struct FS_FILE *entry)
{
    struct FS_FILELIST *lst=entry->Volume->OpenFiles,*prev=NULL;

    while(lst)
    {
        if(lst->UsedCount<FS_FILELIST_BLOCK) {
            lst->File[lst->UsedCount]=entry;
            ++lst->UsedCount;
            return FS_OK;
        }
        prev=lst;
        lst=lst->Next;
    }

    // NO SPACE ON THE LIST, ALLOCATE NEW CHUNK

    lst=(struct FS_FILELIST *)FSMalloc(sizeof(struct FS_FILELIST));

    if(!lst) return FS_NOMEM;

    FSMemset(lst,0,sizeof(struct FS_FILELIST));

    lst->File[0]=entry;
    ++lst->UsedCount;

    if(prev) prev->Next=lst;
    if(!entry->Volume->OpenFiles) entry->Volume->OpenFiles=lst;

    return FS_OK;

}

int FSRemoveFromOpenFileList(struct FS_FILE *entry)
{
    struct FS_FILELIST *lst=entry->Volume->OpenFiles,*prev=NULL;
    int f;

    while(lst)
    {
        for(f=0;f<lst->UsedCount;++f)
        {
            if(lst->File[f]==entry) {
                // FOUND ENTRY, REMOVE
                --lst->UsedCount;

                if(!lst->UsedCount) {
                    // WHOLE LIST BLOCK CAN BE REMOVED
                    if(prev) prev->Next=lst->Next;
                    if(entry->Volume->OpenFiles==lst) entry->Volume->OpenFiles=lst->Next;
                    FSFree(lst);
                    return FS_OK;
                }
                // THERE ARE OTHER ENTRIES, PACK THEM
                FSMemcpy(&(lst->File[f]),&(lst->File[f+1]),(lst->UsedCount-f)*sizeof(struct FS_FILE *));
                lst->File[lst->UsedCount]=NULL;

                return FS_OK;

            }
        }

        prev=lst;
        lst=lst->Next;
    }

    // FILE WAS NOT ON THE LIST, BUT IT'S OK

    return FS_OK;

}





// CREATE A NEW FILE

int FSCreate(clfs_char *name,clfs_32bit attr,struct FS_FILE **fileptr,struct FS_VOLUME *fs)
{
    clfs_char *nptr;
    struct FS_FILE *dir,*file;
    struct FS_FILE *cr;
    clfs_char *path=NULL;
    int error;
    clfs_32bit numentries;
    int ntype;
#ifdef FSCONFIG_WINDOWS_COMPAT
    const clfs_char separators[]="\\/";
#else
    const clfs_char separators[]="/";
#endif



    *fileptr=NULL;

    if(fs==NULL) return FS_NODISK;

    if(fs->Flags&FSATTR_IMMUTABLE) return FS_CANTWRITE;

    //printf("Create init\n");
    ntype=FSGetNameType(name);
    //printf("ntype=%02X\n",ntype);

    if(ntype<0) return FS_BADNAME;

    if((ntype&FSNAME_ENDSLASH) || (ntype&FSNAME_EMPTY)) return FS_BADNAME;

    nptr=name;
    // OBTAIN DIRECTORY
    if(ntype&FSNAME_HASPATH) {
        nptr=__fsfindcharrev(name,NULL,separators);

        // nptr CAN'T BE NULL HERE, SO NO NEED TO CHECK
        path=(clfs_char *)FSMalloc((int)(nptr-name)+2);
        if(!path) return FS_NOMEM;
        // EXTRACT DRIVE/PATH
        FSMemcpy(path,name,(int) (nptr-name+1));
        path[(int)(nptr-name)+1]=0;
        ++nptr;
        //printf("path=%s\n",path);
        dir=(struct FS_FILE *)FSMalloc(sizeof(struct FS_FILE));
        if(!dir) { FSFree(path); return FS_NOMEM; }

        error=FSFindFile(path,dir,fs,FS_TRUE);
        if(error!=FS_OK)
        {
            if(error==FS_OPENDIR) {
                // GET DIRECTORY ADDRESS
                file=dir;
                dir=dir->Dir;
                FSFree(file);
            }
            else {
                FSFree(dir);
                FSFree(path);
                return error;
            }
        }
        //printf("dir=%s\n",(dir->Name)? dir->Name: "\\");
        FSFree(path);
    }
    else {
        // GET CURRENT DIRECTORY
        dir=fs->CurrentDir;
    }


    if(!(dir->F.Attributes&FSATTR_DIR)) {
        // GIVEN PATH IS A FILE, NOT A DIR
        while(dir!=NULL) dir=FSFreeFile(dir);
        return FS_BADNAME;
    }
    // dir HERE IS THE DIRECTORY WHERE FILE WILL BE CREATED



    cr=(struct FS_FILE *)FSMalloc(sizeof(struct FS_FILE));
    if(!cr) {
        while(dir!=NULL) dir=FSFreeFile(dir);
        return FS_NOMEM;
    }




    //printf("find entry\n");
    error=FSFindEntry(nptr,(dir->F.Attributes&FSATTR_DIRCASEINSENS)|(attr&FSATTR_CASEINSENS),cr,dir);

    if((error!=FS_OK)&&(error !=FS_NOTFOUND)) {
        FSFree(cr);
        while(dir!=NULL) dir=FSFreeFile(dir);
        return error;
    }

    if(error!=FS_NOTFOUND) {
        FSFreeEntry(cr);
        //printf("free ok\n");
        while(dir!=NULL) dir=FSFreeFile(dir);
        //printf("dir free ok\n");
        return FS_EXIST;
    }

    // FILE DOESN'T EXIST, CREATE ENTRY

    // COMPLETE MISSING FIELDS IN ENTRY STRUCTURE

    cr->Mode=FSMODE_WRITE;
    cr->F.Attributes=attr;
    cr->Chain=NULL;
    cr->Metadata=NULL;
    cr->Name=NULL;
    clfs_64bit time=FSTimeDateNow();
    cr->F.CreateDate=cr->F.ModifiedDate=time>>16;
    cr->F.ExtendedDates=(time&0xffff)|((time&0xffff)<<16);
    cr->CurrentOffset=0;
    cr->Dir=dir;
    cr->F.EntryType=DE_FILE;
    cr->F.FileSize=0;
    cr->F.StartCluster=0;
    cr->F.NumClusters=0;
    cr->Buffer.Data=NULL;
    cr->Buffer.Used=0;
    cr->Volume=fs;
    error=FSRenameDirEntry(cr,nptr);
    if(error!=FS_OK)
    {
        FSFreeEntry(cr);
        //printf("free ok\n");
        while(dir!=NULL) dir=FSFreeFile(dir);
        //printf("dir free ok\n");
        return error;

    }
    cr->DirEntryCount=FSCalcDirEntryCount(cr,NULL);
    // ONLY INFORMATION MISSING IS DirEntryOffset
    error=FSAllocDirEntries(dir,numentries,&cr->DirEntryOffset);
    if(error!=FS_OK)
    {
    FSFreeEntry(cr);
    //printf("free ok\n");
    while(dir!=NULL) dir=FSFreeFile(dir);
    //printf("dir free ok\n");
    return error;
    }

    // WRITE THE NEW DIRECTORY ENTRY

    error=FSUpdateDirEntry(cr,NULL);
    if(error!=FS_OK)
    {
    FSFreeEntry(cr);
    //printf("free ok\n");
    while(dir!=NULL) dir=FSFreeFile(dir);
    //printf("dir free ok\n");
    return error;
    }

    error=FSAddToOpenFileList(cr);
    if(error!=FS_OK)
    {
    FSFreeEntry(cr);
    //printf("free ok\n");
    while(dir!=NULL) dir=FSFreeFile(dir);
    //printf("dir free ok\n");
    return error;
    }

    *fileptr=cr;


    return FS_OK;

}


// UPDATE CACHED DATA STRUCTURES IF A FILE ENTRY MOVED

void FSDirEntryMoved(struct FS_FILE *dir,clfs_64bit OrgOffset,clfs_64bit EndOrgOffset, clfs_64bit NewOffset)
{
struct FS_FILE *ff;
struct FS_FILELIST *lst;
struct FS_VOLUME *fs=dir->Volume;
int f;

ff=fs->CurrentDir;
while(ff!=NULL) {
if(ff->Dir==dir) {
    if( (ff->DirEntryOffset>=OrgOffset)||(ff->DirEntryOffset<EndOrgOffset)) { ff->DirEntryOffset+=NewOffset-OrgOffset; break; }
}
ff=ff->Dir;
}

lst=fs->OpenFiles;
while(lst) {
for(f=0;f<lst->UsedCount;++f)
{
ff=fs->OpenFiles->File[f];
while(ff!=NULL) {
    if(ff->Dir==dir) {
        if( (ff->DirEntryOffset>=OrgOffset)||(ff->DirEntryOffset<EndOrgOffset)) { ff->DirEntryOffset+=NewOffset-OrgOffset; break; }
    }
ff=ff->Dir;
}
}
lst=lst->Next;
}

return;

}




// FIND UNUSED SPACE ROOM AT GIVEN DIRECTORY TO CREATE numentries ENTRIES
// STORE THE OFFSET OF THE SPACE IN THE GIVEN VARIABLE.
// GROWS DIRECTORY IF NECESSARY

int FSAllocDirEntries(struct FS_FILE *dir,clfs_32bit numentries,clfs_64bit *DirEntryOffset)
{

        // SIMPLE STRATEGY: ALLOCATE FROM END OF DIRECTORY
        // RETURN END-OF-FILE OFFSET SO THE DIRECTORY GROWS

       *DirEntryOffset=dir->F.FileSize;
        return FS_OK;
}

clfs_32bit FSGetDirEntryRecordCount(struct FS_FILE *dir,clfs_64bit DirEntryOffset)
{
    int error;
    clfs_32bit count=0;
    struct FS_FILERECORD *fr,*end;

    error=FSInitFileBuffers(dir);
    if(error!=FS_OK) return error;
    end=dir->Buffer.Data+dir->Volume->ClusterSize;

    fr=FSBufferMap(dir,DirEntryOffset);
    while(fr) {
        if ( (fr->EntryType!=DE_EMPTY)&&(fr->EntryType<PRIMARY_ENTRY_LIMIT) )
        {
            if(count) break;
            ++count;
        }
        DirEntryOffset+=DIRENTRY_SIZE;
        if(DirEntryOffset>=dir->F.FileSize) break;
        ++fr;
        if(fr>=end) fr=FSBufferMap(dir,DirEntryOffset);
    }

    return count;


}


// SKIPS THE CURRENT ENTRY, RETURNS OFFSET OF NEXT PRIMARY RECORD

clfs_64bit FSSkipDirEntry(struct FS_FILE *dir,clfs_64bit DirEntryOffset)
{
    int error;
    struct FS_FILERECORD *fr,*end;

// BUFFERS MUST BE INITIALIZED SOMEWHERE ELSE
//    error=FSInitFileBuffers(dir);
//    if(error!=FS_OK) return error;
    end=dir->Buffer.Data+dir->Volume->ClusterSize;
    if(DirEntryOffset<0) return FS_ERROR;
    if(DirEntryOffset>=dir->F.FileSize) return FS_EOF;
    DirEntryOffset+=DIRENTRY_SIZE;
    fr=FSBufferMap(dir,DirEntryOffset);
    while(fr) {
        if ( (fr->EntryType<PRIMARY_ENTRY_LIMIT) )  return DirEntryOffset;
        DirEntryOffset+=DIRENTRY_SIZE;
        if(DirEntryOffset>=dir->F.FileSize) return dir->F.FileSize;
        ++fr;
        if(fr>=end) fr=FSBufferMap(dir,DirEntryOffset);
    }

    // THIS CAN ONLY HAPPEN WITH SEVERE DATA CORRUPTION
    return FS_ERROR;
}


// REVERSE SKIP DIR ENTRY: FINE THE OFFSET OF PREVIOUS PRIMARY RECORD

clfs_64bit FSRSkipDirEntry(struct FS_FILE *dir,clfs_64bit DirEntryOffset)
{
    int error;
    struct FS_FILERECORD *fr;

    if(DirEntryOffset>dir->F.FileSize) DirEntryOffset=dir->F.FileSize;

// BUFFERS MUST BE INITIALIZED SOMEWHERE ELSE
//    error=FSInitFileBuffers(dir);
//    if(error!=FS_OK) return error;

    DirEntryOffset-=DIRENTRY_SIZE;
    if(DirEntryOffset<0) return FS_EOF; // LAST ENTRY ALREADY FOUND
    fr=FSBufferMap(dir,DirEntryOffset);
    while(fr) {
        if ( (fr->EntryType<PRIMARY_ENTRY_LIMIT) )  return DirEntryOffset;
        DirEntryOffset-=DIRENTRY_SIZE;
        if(DirEntryOffset<0) return FS_EOF; // LAST ENTRY
        --fr;
        if(fr<dir->Buffer.Data) fr=FSBufferMap(dir,DirEntryOffset);
    }

    // THIS CAN ONLY HAPPEN WITH SEVERE DATA CORRUPTION
    return FS_ERROR;
}





// MOVE AN EXISTING DIRECTORY ENTRY TO THE END OF THE DIRECTORY
int FSMoveDirEntryToEnd(struct FS_FILE *dir,clfs_64bit DirEntryOffset)
{
    clfs_32bit numentries=FSGetDirEntryRecordCount(dir,DirEntryOffset);
    clfs_64bit bytesread,newoff;
    clfs_byte *buf;
    if(numentries<=0) return FS_OK; // NO ENTRIES = NOTHING TO DO

    buf=FSMalloc(numentries*DIRENTRY_SIZE);
    if(buf==NULL) return FS_NOMEM;

    FSSeek(dir,DirEntryOffset,SEEK_SET);
    bytesread=FSReadLL(buf,numentries*DIRENTRY_SIZE,dir);
    if(bytesread!=numentries*DIRENTRY_SIZE) { FSFree(buf); return FS_CANTREAD; }

    FSSeek(dir,0,SEEK_END);
    newoff=dir->CurrentOffset;

    bytesread=FSWriteLL(buf,numentries*DIRENTRY_SIZE,dir);
    if(bytesread!=numentries*DIRENTRY_SIZE) { FSFree(buf); return FS_CANTWRITE; }

    FSMemset(buf,0,numentries*DIRENTRY_SIZE);

    FSSeek(dir,DirEntryOffset,SEEK_SET);

    bytesread=FSWriteLL(buf,numentries*DIRENTRY_SIZE,dir);
    if(bytesread!=numentries*DIRENTRY_SIZE) { FSFree(buf); return FS_CANTWRITE; }

    // UPDATE IF ENTRY WAS USED BY AN OPEN FILE

    FSDirEntryMoved(dir,DirEntryOffset,DirEntryOffset+numentries*DIRENTRY_SIZE,newoff);

    return FS_OK;
}


// MAKE ROOM AT GIVEN DIRECTORY TO CREATE numentries ENTRIES
// AT THE GIVEN OFFSET, BY MOVING ENTRIES THAT ARE ABOVE THE GIVEN ONE
// GROWS DIRECTORY IF NECESSARY
// DOES NOT DELETE OR MOVE THE CURRENT ENTRY
// INTENDED TO BE USED WITH THE '.' ENTRY

int FSAllocUnmovableDirEntries(struct FS_FILE *dir,clfs_32bit numentries,clfs_64bit DirEntryOffset)
{
struct FS_VOLUME *fs=dir->Volume;
clfs_32bit havenum=0;
clfs_64bit StartOff,Offset,bytesread,NewOff;
struct FS_FILERECORD *fr;

StartOff=Offset=FSSkipDirEntry(dir,DirEntryOffset);
NewOff=dir->F.FileSize;
if(Offset<0) return FS_INVALIDDATA;
if(Offset==NewOff) return FS_OK;
havenum=Offset-DirEntryOffset;
while(havenum<numentries*DIRENTRY_SIZE)
{
    if(Offset>=dir->F.FileSize) {
        // REACHED END-OF-FILE AND STILL DON'T HAVE ENOUGH ROOM
        // NEED TO GROW DIRECTORY
        int error=FSGrowFile(DirEntryOffset+numentries*DIRENTRY_SIZE,dir);
        if(error!=FS_OK) return error;

        // WARNING: THIS LEAVES ENTRIES WITH RANDOM DATA AWAITING TO BE OVERWRITTEN BY THE NEW UNMOVABLE ENTRY
        NewOff=dir->F.FileSize=DirEntryOffset+numentries*DIRENTRY_SIZE;

        break;
    }
    Offset=FSSkipDirEntry(dir,Offset);
    if(Offset<0) return FS_INVALIDDATA;
    havenum=Offset-DirEntryOffset;

}

clfs_byte *buf=FSMalloc(Offset-StartOff);
if(buf==NULL) return FS_NOMEM;


// READ ALL ENTRIES TO MOVE
FSSeek(dir,StartOff,SEEK_SET);
bytesread=FSReadLL(buf,Offset-StartOff,dir);
if(bytesread!=Offset-StartOff) {
    FSFree(buf);
    return FS_CANTREAD;
}

// WRITE THEM AT END OF DIR
FSSeek(dir,0,SEEK_END);
bytesread=FSWriteLL(buf,Offset-StartOff,dir);
if(bytesread!=Offset-StartOff) {
    FSFree(buf);
    return FS_CANTWRITE;
}

// ERASE THEM FROM ORIGINAL PLACE

// NO NEED TO ERASE SINCE IT WILL BE OVERWRITTEN WITH NEW INFO?
FSMemset(buf,0,Offset-StartOff);
FSSeek(dir,StartOff,SEEK_SET);
bytesread=FSWriteLL(buf,Offset-StartOff,dir);
if(bytesread!=Offset-StartOff) {
    FSFree(buf);
    return FS_CANTWRITE;
}
FSFree(buf);

// NOW UPDATE ALL ENTRIES THAT WERE OPEN TO THEIR NEW OFFSETS
FSDirEntryMoved(dir,StartOff,Offset,NewOff);

return FS_OK;

}


// CHANGE THE NAME OF AN ENTRY

int FSRenameDirEntry(struct FS_FILE *file,clfs_char *newname)
{
    int ntype;
    clfs_char *nname;
    struct FS_RECORDLIST *newmeta,*ptr;

    // VERIFY IF THE NAME IS A VALID FILE NAME
    ntype=FSGetNameType(newname);
    if(ntype<0) return ntype;
    if(ntype&FSNAME_HASPATH) return FS_BADNAME;
    if(ntype&FSNAME_EMPTY) return FS_BADNAME;


    nname=FSMalloc(__fsstrlen(newname)+1);
    if(!nname) return FS_NOMEM;

    int numentries=__fsstrlen(newname);
    numentries-=DIRENTRY_NAME_LENGTH;
    numentries+=NAMERECORD_LENGTH-1;
    numentries/=NAMERECORD_LENGTH;

    int k;
    newmeta=NULL;
    for(k=0;k<numentries;++k) {
        ptr=FSMalloc(sizeof(struct FS_RECORDLIST));
        if(!ptr) {
            while(newmeta) { ptr=newmeta; newmeta=newmeta->Next; FSFree(ptr); }
            FSFree(nname);
            return FS_NOMEM;
        }
        // INSERT AT BEGINNING OF LIST
        ptr->Next=newmeta;
        newmeta=ptr;
    }

    // NOW THAT ALL MEMORY WAS ALLOCATED, MODIFY THE EXISTING ENTRY

    FSMemcpy(nname,newname,__fsstrlen(newname)+1);

    if(file->Name) FSFree(file->Name);
    file->Name=nname;

    // REMOVE ALL NAMERECORDS FROM THE EXISTING METADATA

    struct FS_RECORDLIST *prev;
    prev=ptr=file->Metadata;

    while(ptr!=NULL) {
        if(ptr->Namerec.EntryType==DE_NAME) {
            // REMOVE FROM LIST
            if(ptr==file->Metadata) { file->Metadata=ptr->Next; FSFree(ptr); prev=ptr=file->Metadata; continue; }
            prev->Next=ptr->Next;
            FSFree(ptr);
            ptr=prev->Next;
            continue;
        }
        // SKIP THIS RECORD
        prev=ptr;
        ptr=ptr->Next;
    }


    // UPDATE FILE NAME

    struct FS_NAMERECORD *nextentry;
    clfs_char *namedest;
    nextentry=(struct FS_NAMERECORD *)(&(file->F)+1);

    // UPDATE NAME CRC
    file->F.NameCRC=FSCalcNameCRC(nname);
    file->F.FoldedCRC=FSCalcFoldedNameCRC(nname);

    namedest=file->F.Name;
    // WRITE TO FILERECORD
    while(namedest!=(clfs_char *)nextentry) {
        if(*nname) *namedest++=*nname++;
        else *namedest++=0;
    }

    // CREATE EXTRA METADATA AS NEEDED
        ptr=newmeta;

    int f;
    for(f=0;f<numentries;++f)
    {

        nextentry=&(ptr->Namerec);
        namedest=nextentry->Name;
        nextentry->EntryType=DE_NAME;
        ++nextentry;

    while(namedest!=(clfs_char *)nextentry) {
        if(*nname) *namedest++=*nname++;
        else *namedest++=0;
    }

    ptr=ptr->Next;


    }

    FSAppendMetadata(file,newmeta);

    // MARK ENTRY FOR UPDATE
    file->Mode|=FSMODE_DIRTYDIR;

    return FS_OK;

}

// RECALCULATE NUMBER OF DIRECTORY ENTRIES
clfs_32bit FSCalcDirEntryCount(struct FS_FILE *file,int *__singlefrag)
{
    clfs_32bit numentries;
    struct FS_RECORDLIST *ptr;
    int singlefrag;



    numentries=1; // ONE FOR THE FILERECORD

    // COUNT ENTRIES NEEDED FOR FRAGMENTS
    ptr=file->Chain;
    while(ptr) { ++numentries; ptr=ptr->Next; }

    // CHECK IF THE FILE IS A SINGLE FRAGMENT
    // IF IT IS, THEN WE DON'T NEED AN EXTRA ENTRY
    singlefrag=0;

    if(file->Chain) {
        if(file->Chain->Next==NULL) {
            int f;
            for(f=0;f<DIRENTRY_NUM_FRAGMENTS;++f)
            {
                if(file->Chain->Fragrec.Fragments[f].NumClusters) ++singlefrag;
            }
            if(singlefrag==1) {
                // IT'S A SINGLE-FRAGMENT
                // UPDATE START CLUSTER AND NUMCLUSTER
                file->F.StartCluster=file->Chain->Fragrec.Fragments[0].StartCluster;
                file->F.NumClusters=file->Chain->Fragrec.Fragments[0].NumClusters;

                --numentries;
            }
            else singlefrag=0;



        }
    } else singlefrag=1;




    // COUNT ENTRIES NEEDED FOR METADATA
    // SKIP NAME RECORDS WHEN ENTRY IS A DIRECTORY
    ptr=file->Metadata;
    while(ptr) { if(!((file->F.Attributes&FSATTR_DIR)&&(ptr->Filerec.EntryType==DE_NAME))) ++numentries; ptr=ptr->Next; }


    if(__singlefrag) *__singlefrag=singlefrag;
    return numentries;
}


// UPDATE MINOR ITEMS LIKE ATTRIBUTES OR MODIFIED DATE
// CAN ALSO BE USED TO CHANGE ENTRY TO A DELETED TYPE
// DO NOT USE IF CLUSTER CHAIN CHANGED, NAME CHANGED OR IF METADATA CHANGED

int FSTouchDirEntry(struct FS_FILE *file)
{
    struct FS_FILE *dir=file->Dir;
    int error;
    struct FS_FILERECORD *ptr;

    if(!dir) return FS_OK;  // FILE DOESN'T BELONG TO ANY DIRECTORY, NOTHING TO DO

    ptr=(struct FS_FILERECORD *)FSBufferMap(dir,file->DirEntryOffset);

    if(!ptr) return FS_ERROR;

    ptr->Attributes=file->F.Attributes;
    ptr->CreateDate=file->F.CreateDate;
    ptr->EntryType=file->F.EntryType;
    ptr->ExtendedDates=file->F.ExtendedDates;
    ptr->ModifiedDate=file->F.ModifiedDate;

    FSBufferMarkDirty(dir);

    return FS_OK;
}




// WRITE DIRECTORY ENTRY TO PARENT DIR
// CHECK IF ENOUGH SPACE
// REPLACE THE FILERECORD WITH THE INFORMATION GIVEN IN newinfo IF NOT NULL
// ONLY NEEDED FOR OPEN FILES WHERE THE CLUSTER CHAIN OR THE METADATA NEEDS TO BE UPDATED
// USE TOUCH TO UPDATE SIMPLE FIELDS LIKE ATTRIBUTES
int FSUpdateDirEntry(struct FS_FILE *file,struct FS_FILERECORD *newinfo)
{
    struct FS_FILE *dir;

    clfs_64bit bytesread;
    struct FS_FILERECORD *newentry;
    struct FS_NAMERECORD *nameentry;
    struct FS_FRAGRECORD *fragentry;
    struct FS_METARECORD *metaentry;
    struct FS_RECORDLIST *ptr;

    clfs_32bit f,singlefrag;
    clfs_char *nameptr,*namedest;
    clfs_32bit nameentries,numentries,namelen;
    clfs_32bit dentrycount;
    clfs_64bit dentryoffset;
    int error;


        if(file->Name) namelen=__fsstrlen(file->Name);
        else namelen=0;

        if(file->F.Attributes&FSATTR_DIR) {
            dir=file; // DIRECTORIES ARE UPDATED ONLY AT THEIR OWN RECORD
            dentryoffset=file->DirEntryOffset2;
            dentrycount=file->DirEntryCount2;
        }
        else {
            dir=file->Dir;
            dentryoffset=file->DirEntryOffset;
            dentrycount=file->DirEntryCount;
        }

        // COUNT ENTRIES FOR RECORD AND NAME

        numentries=FSCalcDirEntryCount(file,&singlefrag);

        newentry=(struct FS_FILERECORD *)FSMalloc(((numentries>dentrycount)? numentries:dentrycount)*sizeof(struct FS_FILERECORD));

        if(!newentry) return FS_NOMEM;


        if(numentries>dentrycount) {
            // NEED TO REALLOCATE ENTRY WITHIN DIRECTORY

            if(dir==file) {
             // A DIRECTORY NEEDS TO UPDATE THE '.' ENTRY, MAKE ROOM BY MOVING OTHER FILES
             error=FSAllocUnmovableDirEntries(dir,numentries,dentryoffset);
                if(error!=FS_OK) {
                    FSFree(newentry);
                    return error;
                }


            }
            else {
            // MARK ALL ENTRIES AS FREE, THEN ALLOCATE NEW ONES
            FSMemset(newentry,0,dentrycount*sizeof(struct FS_FILERECORD));
            FSSeek(dir,dentryoffset,SEEK_SET);
            FSWriteLL((clfs_byte *)newentry,dentrycount*sizeof(struct FS_FILERECORD),dir);

            // NOW ALLOCATE NEW ENTRIES
            error=FSAllocDirEntries(dir,numentries,&dentryoffset);
            if(error!=FS_OK) {
                FSFree(newentry);
                return error;
            }
            file->DirEntryOffset=dentryoffset;
            }

            // FINALLY MARK THE DIRECTORY AS DIRTY
            dir->Mode|=FSMODE_DIRTYDIR;

        }




        // UPDATE FILE RECORD

        if(newinfo) {
            FSMemcpy(newentry,newinfo,sizeof(struct FS_FILERECORD)-DIRENTRY_NAME_LENGTH);
            // GUARANTEE THAT THE NAME AND START CLUSTERS ARE TAKEN FROM THE REAL INFO FILE
            FSMemcpy(newentry->Name,file->F.Name,DIRENTRY_NAME_LENGTH);

            newentry->NumClusters=file->F.NumClusters;
            newentry->StartCluster=file->F.StartCluster;
        }
        else FSMemcpy(newentry,&file->F,sizeof(struct FS_FILERECORD));

        // ADD FRAGMENT RECORDS


        fragentry=(struct FS_FRAGRECORD *)(newentry+1);

        if(!singlefrag) {

        ptr=file->Chain;
        while(ptr) {
            FSMemcpy(fragentry,&ptr->Fragrec,sizeof(struct FS_FRAGRECORD));
            ++fragentry;
            ptr=ptr->Next;
        }

        }

        // ADD METADATA RECORDS

        metaentry=(struct FS_METARECORD *)fragentry;

        ptr=file->Metadata;
        while(ptr) {
            if( !((file->F.Attributes&FSATTR_DIR)&&(ptr->Metarec.EntryType==DE_NAME)))
            {
            FSMemcpy(metaentry,&ptr->Fragrec,sizeof(struct FS_METARECORD));
            ++metaentry;
            }
            ptr=ptr->Next;
        }

        if(file->F.Attributes&FSATTR_DIR) file->DirEntryCount2=numentries;
        else file->DirEntryCount=numentries;

        // DONE, newentry CONTAINS DATA READY TO BE WRITTEN TO DISK
        if(numentries<dentrycount) {
            // FILL EMPTY ENTRIES AFTER
            FSMemset(metaentry,0,(dentrycount-numentries)*sizeof(struct FS_FILERECORD));
            numentries=dentrycount;
        }


        FSSeek(dir,dentryoffset,SEEK_SET);
        bytesread=FSWriteLL((clfs_byte *)newentry,numentries*sizeof(struct FS_FILERECORD),dir);
        FSFree(newentry);

        if(bytesread!=numentries*sizeof(struct FS_FILERECORD)) {
            // WE FAILED TO WRITE THE DIRECTORY ENTRY
            return FS_CANTWRITE;
        }
        return FS_OK;

}

void FSFreeAllClusters(struct FS_FILE *file)
{
    // RELEASE ALL CLUSTERS, MARK AS FREE
    struct FS_RECORDLIST *ptr;
    int k,error;

    ptr=file->Chain;
    while(ptr) {
        for(k=0;k<DIRENTRY_NUM_FRAGMENTS;++k)
        {
            if(ptr->Fragrec.Fragments[k].NumClusters) {
            error=FSMarkClustersFree(ptr->Fragrec.Fragments[k].StartCluster,ptr->Fragrec.Fragments[k].NumClusters,file->Volume);
            //if(error!=FS_OK) { } // DO NOTHING ON ERROR, WORST CASE THERE WILL BE LOST CLUSTERS

            }
        }
        ptr=ptr->Next;
    }

}


int FSClose(struct FS_FILE *file)
{
int error;
struct FS_FILE *dir;

if(file==NULL) return FS_ERROR;

//printf("close buffers\n");
error=FSBufferFlush(file);
if(error!=FS_OK) return error;

if((file->F.EntryType==DE_DELETED) | (file->Mode&FSMODE_WRITE)) {
    error=FSUpdateDirEntry(file,NULL);
    if(error!=FS_OK) return error;
}
        if(file->F.EntryType==DE_DELETED) {
            FSFreeAllClusters(file);
        }

FSRemoveFromOpenFileList(file);

dir=FSFreeFile(file);

while(dir) dir=FSFreeFile(dir);

return FS_OK;
}


int FSMarkClustersUsed(clfs_64bit Cluster,clfs_64bit NumClusters,struct FS_VOLUME *fs)
{
    // WRITE BITS IN THE FREE CLUSTERS BITMAP

    clfs_64bit Offset;
    clfs_32bit *ptrleft,*ptrright;
    clfs_32bit maskleft,maskright;
    int error;

    maskleft=1<<(Cluster&31);
    maskleft=~(maskleft-1);     // ALL BITS STARTING WITH THE GIVEN CLUSTER
    maskright=1<<((Cluster+NumClusters-1)&31);
    maskright=((maskright-1)<<1)|1;     // ALL BITS UP TO THE GIVEN ONE
    Offset=(Cluster>>3)&(~(fs->ClusterSize-1));
    ptrleft=ptrright=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
    ptrleft+=((Cluster>>3)&(fs->ClusterSize-1))>>2;
    ptrright+=(((Cluster+NumClusters-1)>>3)&(fs->ClusterSize-1))>>2;

    if(ptrright==ptrleft) { maskleft&=maskright; maskright=maskleft; }


    // READ CLUSTER
    error=FSBufferFill(fs->FreeClusters,Offset);
    if(error!=FS_OK) return error;

    *ptrleft|=maskleft;

    FSBufferMarkDirty(fs->FreeClusters);

    ++ptrleft;
    while(ptrleft<ptrright)
    {
        if(ptrleft>=(clfs_32bit *) (fs->FreeClusters->Buffer.Data+fs->ClusterSize)) {
            Offset+=fs->ClusterSize;
            error=FSBufferFill(fs->FreeClusters,Offset);
            if(error!=FS_OK) return error;
            ptrleft=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
            ptrright-=(fs->ClusterSize)>>2;
        }
        *ptrleft=0xFFFFFFFF;
        FSBufferMarkDirty(fs->FreeClusters);

        ++ptrleft;
    }

    if(ptrright>=(clfs_32bit *) (fs->FreeClusters->Buffer.Data+fs->ClusterSize)) {
        Offset+=fs->ClusterSize;
        error=FSBufferFill(fs->FreeClusters,Offset);
        if(error!=FS_OK) return error;
        ptrright-=(fs->ClusterSize)>>2;
    }

    *ptrright|=maskright;
    FSBufferMarkDirty(fs->FreeClusters);

    return FS_OK;

}


int FSMarkClustersFree(clfs_64bit Cluster,clfs_64bit NumClusters,struct FS_VOLUME *fs)
{
    // WRITE BITS IN THE FREE CLUSTERS BITMAP

    clfs_64bit Offset;
    clfs_32bit *ptrleft,*ptrright;
    clfs_32bit maskleft,maskright;
    int error;

    maskleft=1<<(Cluster&31);
    maskleft=~(maskleft-1);     // ALL BITS STARTING WITH THE GIVEN CLUSTER
    maskright=1<<((Cluster+NumClusters-1)&31);
    maskright=((maskright-1)<<1)|1;     // ALL BITS UP TO THE GIVEN ONE
    Offset=(Cluster>>3)&(~(fs->ClusterSize-1));
    ptrleft=ptrright=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
    ptrleft+=((Cluster>>3)&(fs->ClusterSize-1))>>2;
    ptrright+=(((Cluster+NumClusters-1)>>3)&(fs->ClusterSize-1))>>2;

    if(ptrright==ptrleft) { maskleft&=maskright; maskright=maskleft; }


    // READ CLUSTER
    error=FSBufferFill(fs->FreeClusters,Offset);
    if(error!=FS_OK) return error;

    *ptrleft&=~maskleft;

    FSBufferMarkDirty(fs->FreeClusters);

    ++ptrleft;
    while(ptrleft<ptrright)
    {
        if(ptrleft>=(clfs_32bit *) (fs->FreeClusters->Buffer.Data+fs->ClusterSize)) {
            Offset+=fs->ClusterSize;
            error=FSBufferFill(fs->FreeClusters,Offset);
            if(error!=FS_OK) return error;
            ptrleft=(clfs_32bit *)fs->FreeClusters->Buffer.Data;
            ptrright-=(fs->ClusterSize)>>2;
        }
        *ptrleft=0;
        FSBufferMarkDirty(fs->FreeClusters);

        ++ptrleft;
    }

    if(ptrright>=(clfs_32bit *) (fs->FreeClusters->Buffer.Data+fs->ClusterSize)) {
        Offset+=fs->ClusterSize;
        error=FSBufferFill(fs->FreeClusters,Offset);
        if(error!=FS_OK) return error;
        ptrright-=(fs->ClusterSize)>>2;
    }

    *ptrright&=~maskright;
    FSBufferMarkDirty(fs->FreeClusters);

    return FS_OK;

}


// RECURSIVELY SCAN DIRECTORIES AND MARK USED CLUSTERS BY ALL FILES AND ENTRIES, INCLUDING SELF

int FSRescanUsedClusters(struct FS_FILE *Dir)
{
    // DO NOT DECLARE TOO MUCH DATA HERE, SINCE IT'S RECURSIVE

    int error,k;
    struct FS_FILE *entry;
    struct FS_RECORDLIST *ptr;

    // DIRECTORY MUST BE ALREADY OPEN

    FSSeek(Dir,0,SEEK_SET);

    while( (error=FSGetNextDirEntry(Dir,&entry)) !=FS_EOF)
    {
        if(error!=FS_OK) return error;

        // DO NOT PROCESS ANYTHING THAT'S NOT A FILE
        if(entry->F.EntryType!=DE_FILE) {
            while(entry->Chain) { ptr=entry->Chain; entry->Chain=entry->Chain->Next; FSFree(ptr); }
            while(entry->Metadata) { ptr=entry->Metadata; entry->Metadata=entry->Metadata->Next; FSFree(ptr); }
            if(entry->Name) FSFree(entry->Name);
            FSFree(entry);
            continue;
        }

        // GOT ENTRY, ADD FRAGMENTS

        ptr=entry->Chain;
        while(ptr) {
            for(k=0;k<DIRENTRY_NUM_FRAGMENTS;++k)
            {
                if(ptr->Fragrec.Fragments[k].NumClusters) {
                error=FSMarkClustersUsed(ptr->Fragrec.Fragments[k].StartCluster,ptr->Fragrec.Fragments[k].NumClusters,Dir->Volume);
                if(error!=FS_OK) {
                    while(entry->Chain) { ptr=entry->Chain; entry->Chain=entry->Chain->Next; FSFree(ptr); }
                    while(entry->Metadata) { ptr=entry->Metadata; entry->Metadata=entry->Metadata->Next; FSFree(ptr); }
                    if(entry->Name) FSFree(entry->Name);
                    FSFree(entry);
                    return error;
                }
                }
            }
            ptr=ptr->Next;
        }

        if( (entry->F.Attributes&FSATTR_DIR) && !((entry->F.Name[0]=='.') && (entry->F.Name[1]==0))) {
            // CLEANUP ANY EXTRA DATA
            while(entry->Chain) { ptr=entry->Chain; entry->Chain=entry->Chain->Next; FSFree(ptr); }
            while(entry->Metadata) { ptr=entry->Metadata; entry->Metadata=entry->Metadata->Next; FSFree(ptr); }
            if(entry->Name) { FSFree(entry->Name); entry->Name=NULL; }

            // PASS THE CURRENT ENTRY TO BE OVERWRITTEN
            error=FSOpenDirLL(entry,entry->Volume,entry->F.StartCluster);
            if(error!=FS_OK) {
                FSFree(entry);
                return error;
            }
            // NOW THE DIRECTORY IS READY TO BE USED
            // RECURSE
            error=FSRescanUsedClusters(entry);
            if(error!=FS_OK) {
                while(entry->Chain) { ptr=entry->Chain; entry->Chain=entry->Chain->Next; FSFree(ptr); }
                while(entry->Metadata) { ptr=entry->Metadata; entry->Metadata=entry->Metadata->Next; FSFree(ptr); }
                // NO NEED TO FREE Name, SINCE IT IS NOT FILLED BY FSOpenDirLL
                FSFree(entry);
                return error;
            }
        }
        while(entry->Chain) { ptr=entry->Chain; entry->Chain=entry->Chain->Next; FSFree(ptr); }
        while(entry->Metadata) { ptr=entry->Metadata; entry->Metadata=entry->Metadata->Next; FSFree(ptr); }
        if(entry->Name) FSFree(entry->Name);
        FSFree(entry);
    }

return FS_OK;
}


// CREATE OR RE-CREATE THE FREE CLUSTERS BITMAP ON A READ-ONLY VOLUME

int FSRescanFreeSpace(struct FS_VOLUME *fs)
{
    clfs_64bit byteswritten;
    clfs_64bit clusternum;
    int error;

    if(!fs->FreeClusters) {
    // MUST OPEN THE BITMAP FIRST
    error=FSOpen(FSSTRING_FREECLUSTERS,FSMODE_READ,&fs->FreeClusters,fs );
    if(error!=FS_OK) {
            // BITMAP DOESN'T EXIST, MUST BE CREATED FIRST
        return FS_ERROR;
    }

    }
    // FORCE A MODE CHANGE IN THE BITMAP
    fs->FreeClusters->Mode=FSMODE_MODIFY|FSMODE_WRITE|FSMODE_NOGROW;
    // AND ALLOW WRITING TO THE VOLUME
    fs->Flags&=~FSATTR_IMMUTABLE;

    // MARK ENTIRE DISK SPACE AS FREE

    error=FSInitFileBuffers(fs->FreeClusters);
    if(error!=FS_OK) {
        FSClose(fs->FreeClusters);
        fs->FreeClusters=NULL;
        fs->Flags|=FSATTR_IMMUTABLE;
        return error;
    }

    // WIPEOUT ENTIRE BITMAP
    fs->FreeClusters->Buffer.Offset=0;
    FSMemset(fs->FreeClusters->Buffer.Data,0,fs->ClusterSize);
    while(fs->FreeClusters->Buffer.Offset<fs->FreeClusters->F.FileSize) {
        clusternum=FSOffsetToCluster(fs->FreeClusters->Buffer.Offset,fs->FreeClusters);
        if(clusternum==-1) {
           // DAMAGED ALLOCATION CHAIN ON THE BITMAP, PANIC AND EXIT
            FSClose(fs->FreeClusters);
            fs->FreeClusters=NULL;
            fs->Flags|=FSATTR_IMMUTABLE;
            return FS_CANTWRITE;
        }
        byteswritten=FSWriteCluster(clusternum,fs->FreeClusters->Buffer.Data,1, fs);
        if(byteswritten!=fs->ClusterSize) {
            FSClose(fs->FreeClusters);
            fs->FreeClusters=NULL;
            fs->Flags|=FSATTR_IMMUTABLE;
            return FS_CANTWRITE;
        }
        fs->FreeClusters->Buffer.Offset+=fs->ClusterSize;

    }



    // NOW START SCANNING THE ENTIRE VOLUME

    // CLUSTER 0 IN THE VOLUME IS ALWAYS USED, SO IT IS USED AS THE 'DIRTY BIT'
    // CLUSTER 0 IS MARKED AS FREE WHEN THE VOLUME IS UNMOUNTED CORRECTLY, AND
    // MARKED AS USED WHEN THE VOLUME IS MOUNTED
    // IF THE BIT IS 1 WHEN THE VOLUME IS MOUNTED, THEN THE FILE SYSTEM DIDN'T CLOSE PROPERLY AND
    // THE BITMAP COULD BE IN AN INCONSISTENT STATE, MUST BE RESCANNED

    // MAKE BUFFER CONTAIN VALID DATA
    fs->FreeClusters->Buffer.Offset=0;
    fs->FreeClusters->Buffer.Used=FSBUFFER_HASDATA;


    error=FSMarkClustersUsed(0,1,fs);
    if(error!=FS_OK)
    {
        FSClose(fs->FreeClusters);
        fs->FreeClusters=NULL;
        fs->Flags|=FSATTR_IMMUTABLE;
        return FS_CANTWRITE;
    }


    error=FSRescanUsedClusters(&fs->RootDir);
    if(error!=FS_OK)
    {
        FSClose(fs->FreeClusters);
        fs->FreeClusters=NULL;
        fs->Flags|=FSATTR_IMMUTABLE;
        return FS_CANTWRITE;
    }

    FSBufferFlush(fs->FreeClusters);    // ENSURE THAT BUFFERS ARE WRITTEN TO DISK

    return FS_OK;
}


// ALLOCATES A FS_FILE STRUCTURE AND FILLS IT WITH THE INFORMATION OF THE DIRECTORY ENTRY AT CURRENT OFFSET


int FSGetNextDirEntry(struct FS_FILE *Dir,struct FS_FILE **nextentry)
{
    struct FS_FILE *result;
    int error;

    if(Dir==NULL) return FS_ERROR;

    if(Dir->CurrentOffset>=Dir->F.FileSize) return FS_EOF;

    result=FSMalloc(sizeof(struct FS_FILE));
    if(!result) return FS_NOMEM;


    error=FSGetDirEntryData(Dir,Dir->CurrentOffset,result);

    if(error!=FS_OK) { FSFree(result); return error; }

    Dir->CurrentOffset=FSSkipDirEntry(Dir,Dir->CurrentOffset);
    if(Dir->CurrentOffset<0) {
        Dir->CurrentOffset=Dir->F.FileSize;
        FSFreeEntry(result);
        return FS_ERROR;
    }

    *nextentry=result;
    return FS_OK;
}



// USE THE FREE SPACE BITMAP IN THE CURRENT DIRECTORY.
// MUST BE CALLED WITH CURRENT DIR SET TO THE ROOT DIR OR ANY OTHER DIR WITH VALID FREE SPACE BITMAP

int FSMountFreeSpace(struct FS_VOLUME *fs)
{
    clfs_32bit dirtybit=0;
    int error;

    if(!fs->FreeClusters) {
    // MUST OPEN THE BITMAP FIRST
    error=FSOpen(FSSTRING_FREECLUSTERS,FSMODE_READ,&fs->FreeClusters,fs );
    if(error!=FS_OK) return error;
    }



    FSRemoveFromOpenFileList(fs->FreeClusters);
    FSSeek(fs->FreeClusters,0,SEEK_SET);

    error=FSReadLL((clfs_byte *)&dirtybit,1,fs->FreeClusters);
    if(error!=1) {
        FSFreeFile(fs->FreeClusters);
        return FS_CANTREAD;
    }
    // NOW CHECK CLUSTER 0 BIT, USED AS DIRTY
    if(dirtybit&1) {
        // VOLUME WASN'T UNMOUNTED PROPERLY, MUST RESCAN
        return FS_INVALIDDATA;
    }

    // OTHERWISE WE HAVE A HEALTHY FILE-SYSTEM. ENABLE WRITING

    // FORCE A MODE CHANGE IN THE BITMAP
    fs->FreeClusters->Mode=FSMODE_MODIFY|FSMODE_WRITE|FSMODE_NOGROW;
    // AND ALLOW WRITING TO THE VOLUME
    fs->Flags&=~FSATTR_IMMUTABLE;



    error=FSMarkClustersUsed(0,1,fs);     // SET DIRTY BIT BACK TO 1
    if(error!=FS_OK)
    {
        // CAN'T MARK CLUSTERS --> REVERT BACK TO READ-ONLY
        fs->Flags|=FSATTR_IMMUTABLE;
        FSClose(fs->FreeClusters);
        fs->FreeClusters=NULL;
        return error;
    }

    error=FSBufferFlush(fs->FreeClusters);    // AND WRITE DIRTY BIT BACK TO DISK
    if(error!=FS_OK)
    {
        // CAN'T MARK CLUSTERS --> REVERT BACK TO READ-ONLY
        fs->Flags|=FSATTR_IMMUTABLE;
        FSClose(fs->FreeClusters);
        fs->FreeClusters=NULL;
        return error;
    }

    return FS_OK;

}

// UNMOUNT THE GIVEN VOLUME, USER FILES MUST BE CLOSED PRIOR TO CALLING
// RETURN FS_OPENFILE ERROR IF THERE ARE ANY OPEN FILES
// THIS CAN BE OVERRIDDEN WITH forceclose, WHICH WILL TRY TO CLOSE EACH
// FILE BEFORE UNMOUNTING

int FSUnmountVolume(struct FS_VOLUME *fs,clfs_32bit forceclose)
{
    if(fs->OpenFiles && (!forceclose)) return FS_OPENFILE;

    struct FS_FILE *ff;
    struct FS_RECORDLIST *fr;
    struct FS_FILELIST *lst;
    int f;
    int error,errno;

    errno=FS_OK;

    rescan:
    lst=fs->OpenFiles;
    while(lst) {
    for(f=0;f<lst->UsedCount;++f)
    {
    ff=fs->OpenFiles->File[f];
    while(ff!=NULL) {
        error=FSClose(ff);
        if(error!=FS_OK) errno=error;       // RECORD ERRORS BUT DON'T STOP CLOSING
        // THE FILE WAS REMOVED FROM THE LIST, SO THE ENTIRE LIST COULD BE INVALID RIGHT NOW
        goto rescan;
    ff=ff->Dir;
    }
    }
    lst=lst->Next;
    }

    // ALL FILES ARE CLOSED

    // CLOSE THE CURRENT DIR PATH
    if(fs->CurrentDir!=&fs->RootDir) {
        ff=fs->CurrentDir;
        fs->CurrentDir=&fs->RootDir;         // CHANGE TO ROOTDIR
        error=FSClose(ff);
        if(error!=FS_OK) errno=error;       // RECORD ERRORS BUT DON'T STOP CLOSING
    }

    fs->CurrentDir=&fs->RootDir;


    // FINALLY, CLOSE THE ROOT DIRECTORY

    FSUpdateDirEntry(&fs->RootDir,NULL);
    FSBufferFlush(&fs->RootDir);

    // AND LAST, UPDATE THE FREE SPACE BITMAP, CLEAR THE DIRTY BIT
    if(fs->FreeClusters) {

    error=FSMarkClustersFree(0,1,fs);
    if(error!=FS_OK) errno=error;       // RECORD ERRORS BUT DON'T STOP CLOSING

    error=FSClose(fs->FreeClusters);
    if(error!=FS_OK) errno=error;       // RECORD ERRORS BUT DON'T STOP CLOSING
    }

    fs->FreeClusters=NULL;


    ff=&fs->RootDir;
    // RELEASE ALL MEMORY FROM ROOT DIR
    while(ff->Chain) { fr=ff->Chain->Next; FSFree(ff->Chain); ff->Chain=fr; }
    while(ff->Metadata) { fr=ff->Metadata->Next; FSFree(ff->Metadata); ff->Metadata=fr; }
    if(ff->Buffer.Data) FSFree(ff->Buffer.Data);
    if(ff->Name) FSFree(ff->Name);

    // RELEASE ALL MEMORY FROM VOLUME
    if(fs->VolName) FSFree(fs->VolName);

    return errno;
}


int FSMkdir(clfs_char *name,struct FS_VOLUME *fs)
{
    struct FS_FILE *dir;
    struct FS_FILERECORD *fr;
    int error;
    clfs_64bit byteswritten;

    error=FSCreate(name,0,&dir,fs);
    if(error!=FS_OK) return error;

    // DIRECTORY WAS CREATED SUCCESSFULLY
    // NOW WRITE PROPER INFORMATION FOR A DIRECTORY

    // ALLOCATE BUFFER FOR THE DIRECTORY
    error=FSInitFileBuffers(dir);
    if(error!=FS_OK) {
        FSClose(dir);
        return error;
    }

    dir->Buffer.Offset=0;
    dir->Buffer.Used=FSBUFFER_HASDATA|FSBUFFER_DIRTY;
    // CLEAR THE BUFFER FOR SECURITY REASONS
    FSMemset(dir->Buffer.Data,0,fs->ClusterSize);

    // STORE BASIC INFORMATION, NAME AND FIRST CLUSTER WILL BE PROPERLY OVERWRITTEN AFTER THE SPACE IS ALLOCATED
    fr=(struct FS_FILERECORD *)dir->Buffer.Data;
    FSMemcpy(fr,&dir->F,sizeof(struct FS_FILERECORD));
    fr->Attributes|=FSATTR_DIR;
    byteswritten=FSWriteLL(dir->Buffer.Data,DIRENTRY_SIZE,dir);
    if(byteswritten!=DIRENTRY_SIZE) {
        FSClose(dir);
        return FS_CANTWRITE;
    }

    // UPDATE ENTRY AT PARENT DIRECTORY WITH NEWLY ALLOCATED CLUSTER. FORCE CHANGE TO FSATTR_DIR
    error=FSUpdateDirEntry(dir,fr);
    if(error!=FS_OK) {
        // WARNING: IF THIS FAILS THERE WILL BE LOST CLUSTERS
        FSClose(dir);
        return error;

   }

   // SET ATTRIBUTE TO DIRECTORY
   dir->F.Attributes|=FSATTR_DIR;
   // AND CHANGE ACCESS MODE TO MODIFY
   dir->Mode=FSMODE_WRITE|FSMODE_MODIFY;
   dir->DirEntryCount2=1;
   dir->DirEntryOffset2=0;

   // CHANGE RECORD ITEMS TO '.'
   FSMemset(fr->Name,0,DIRENTRY_NAME_LENGTH);
   FSMemset(dir->F.Name,0,DIRENTRY_NAME_LENGTH);
   dir->F.Name[0]=fr->Name[0]='.';
   dir->F.NameCRC=dir->F.FoldedCRC=fr->FoldedCRC=fr->NameCRC=FSCalcNameCRC(".");
   fr->FileSize=64;
   fr->StartCluster=dir->F.StartCluster;
   fr->NumClusters=dir->F.NumClusters;

       // WARNING: IF THIS FAILS THERE WILL BE LOST CLUSTERS
       FSClose(dir);
//       return error;

//  }

   return FS_OK;


}


int FSChdir(clfs_char *name,struct FS_VOLUME *fs)
{
 int error;
 struct FS_FILE *dir,*old;


  error=FSOpenDir(name,&dir,fs);

  if(error!=FS_OK) {
      return error;
  }

  FSRemoveFromOpenFileList(dir);

  old=fs->CurrentDir;
  fs->CurrentDir=dir;
  if(old!=NULL) FSClose(old);
  return FS_OK;


}
// HIGH LEVEL DIRECTORY OPEN FUNCTION
int FSOpenDir(clfs_char *name,struct FS_FILE **fileptr,struct FS_VOLUME *fs)
{
 int error;
 struct FS_FILE *dir,*tmp;

 dir=(struct FS_FILE *) FSMalloc(sizeof(struct FS_FILE));
 if(dir==NULL) return FS_NOMEM;
 tmp=(struct FS_FILE *) FSMalloc(sizeof(struct FS_FILE));
 if(tmp==NULL) { FSFree(dir); return FS_NOMEM; }


 error=FSFindFile(name,dir,fs,FS_TRUE);

  if( (error!=FS_OK)&&(error!=FS_OPENDIR)) {
      FSFree(tmp);
      FSFree(dir);
      return error;
  }

  if(error==FS_OPENDIR)
  {
      // FINDFILE RETURNED THE ALREADY OPEN DIRECTORY POINTER IN THE Dir FIELD
      *fileptr=dir->Dir;
      FSFree(tmp);
      FSFree(dir);
      return FS_OK;
  }

  if(!(dir->F.Attributes&FSATTR_DIR)) {
      // NOT A DIRECTORY
      while(dir!=NULL) dir=FSFreeFile(dir);
      FSFree(tmp);
      return FS_NOTFOUND;
  }

  // FOUND A VALID DIRECTORY, NOW OPEN IT
  error=FSOpenDirLL(tmp,fs,dir->F.StartCluster);

  if(error!=FS_OK) {
      while(dir!=NULL) dir=FSFreeFile(dir);
      FSFree(tmp);
      return error;
  }

  // FSOpenDirLL DOES **NOT** FILL THE FOLLOWING MEMBERS (CALLING FUNCTION IS RESPONSIBLE FOR THAT)
  // Dir = Pointer to the parent directory
  // Name = Name of the directory
  // DirEntryOffset
  // DirEntryCount

  tmp->DirEntryOffset=dir->DirEntryOffset;
  tmp->DirEntryCount=dir->DirEntryCount;


  tmp->Dir=dir->Dir;
  tmp->Name=dir->Name;

  dir->Name=NULL;
  FSFreeEntry(dir);

  error=FSAddToOpenFileList(tmp);
  if(error!=FS_OK)
  {
     while(tmp!=NULL) tmp=FSFreeFile(dir);
     return error;
  }
  *fileptr=tmp;

  return FS_OK;

}



// GET CURRENT WORKING DIRECTORY PATH AS STRING
// THIS FUNCTION ALLOCATES ITS OWN BUFFER OF THE APPROPRIATE SIZE
// NO TRAILING SLASH IS INCLUDED
int FSGetcwd(clfs_char **bufferptr, struct FS_VOLUME *fs)
{
    struct FS_FILE *dir=fs->CurrentDir;
    struct FS_FILE *ptr;
    clfs_32bit pathlen=0;
    clfs_char *path;

    for(ptr=dir;ptr!=&fs->RootDir;ptr=ptr->Dir)
    {
        if(ptr->Name)  pathlen+=__fsstrlen(ptr->Name);
        ++pathlen;
    }

    if(!pathlen) pathlen++;


    path=FSMalloc(pathlen+1);   // ONE EXTRA BYTE FOR NULL TERMINATOR

    if(!path) return FS_NOMEM;

    path[pathlen]=0;

    for(ptr=dir;ptr!=&fs->RootDir;ptr=ptr->Dir)
    {
        if(ptr->Name) {
            pathlen-=__fsstrlen(ptr->Name);
            FSMemcpy(path+pathlen,ptr->Name,__fsstrlen(ptr->Name));
        }
        path[--pathlen]='/';
    }
    if(pathlen) path[0]='/';

    *bufferptr=path;
    return FS_OK;


}

const char * const FSErrorMsg[]={
    "OK",
    "End of File",
    "Invalid Name",
    "Invalid Data",
    "File Not Found",
    "Cannot write to disk",
    "No disk in drive",
    "Device changed",
    "Maximum files exceeded",
    "File already open",
    "Disk full",
    "File already exists",
    "Insufficient memory",
    "Cannot read from disk",
    "Unknown error",
    "File is being used"
};

// RENAME A FILE
int FSRename(clfs_char *oldfile,clfs_char *newname,struct FS_VOLUME *fs)
{
struct FS_FILE *file;
struct FS_FILE entry;
int error;


error=FSFindFile(oldfile,&entry,fs,FS_TRUE);
if((error!=FS_OK)&&(error!=FS_OPENFILE)) {
    return error;
}



if(error==FS_OPENFILE) file=entry.Dir;
else file=&entry;

// ENTRY HAS A VALID ENTRY
error=FSRenameDirEntry(file,newname);

if(file->F.Attributes&FSATTR_DIR)
{
    // FOR DIRECTORIES, UPDATE THE ENTRY IN THE PARENT DIRECTORY
file->F.Attributes&=~FSATTR_DIR;
struct FS_FILERECORD tmp;
FSMemcpy(&tmp,&file->F,sizeof(struct FS_FILERECORD));
tmp.Attributes|=FSATTR_DIR;
error=FSUpdateDirEntry(file,&tmp);
file->F.Attributes|=FSATTR_DIR;
if(file!=&entry) {
    // IF THE DIRECTORY WAS ALREADY OPEN, THE ENTRY SHOULD CONTAIN THE '.' NAME
    // BUT THE NAME FIELD CONTAINS THE ACTUAL NAME, SO WE NEED TO FIX THAT MANUALLY

    clfs_char *realname=file->Name;

    file->Name=NULL;
    // IGNORE ERRORS FROM FSRenameDirEntry SINCE IT'S IN THE MIDDLE OF A CRITICAL OPERATION
    FSRenameDirEntry(file,(clfs_char *)".");

    if(file->Name) FSFree(file->Name);
    file->Name=realname;


}
FSFreeEntryData(&entry);
return error;
}


error=FSUpdateDirEntry(file,NULL);
FSFreeEntryData(&entry);
return error;

}


// MARK AN OPEN FILE FOR DELETION ON CLOSE
void FSDeleteOnClose(struct FS_FILE *file)
{
    file->F.EntryType=DE_DELETED;
    return;
}


// DELETE A FILE FROM THE FILESYSTEM

int FSDelete(clfs_char *name,struct FS_VOLUME *fs)
{
    struct FS_FILE *file;
    int error;

    error=FSOpen(name,FSMODE_APPEND|FSMODE_NOCREATE,&file,fs);
    if(error!=FS_OK) return error;

    FSDeleteOnClose(file);

    error=FSClose(file);
    return error;

}

// ALLOCATE MEMORY AND RETURN A STRING WITH ABSOLUTE PATH OF AN OPEN FILE

clfs_char *FSGetFullPath(struct FS_FILE *file)
{
    int totalname,depth,idx;
    struct FS_FILE *dir=file->Dir;

    depth=0;

    // GET DIRECTORY DEPTH
    dir=file;
    while(dir!=NULL) { ++depth; dir=dir->Dir; }


    // NOW GET NAME INFORMATION FOR ALL PARENT DIRS
    struct pathdata {
        clfs_32bit length;
        clfs_char *string;
    } *pathd;

    pathd=FSMalloc(sizeof(struct pathdata)*(depth+1));
    if(!pathd) return NULL;

    dir=file;
    idx=depth;
    totalname=0;
    while(dir!=NULL) {
        pathd[idx].string=dir->Name;
        pathd[idx].length=(dir->Name)? __fsstrlen(dir->Name):0;
        totalname+=pathd[idx].length+1;
        dir=dir->Dir;
        --idx;
    }
    clfs_char *path=FSMalloc(totalname);  // ALREADY INCLUDES SPACE FOR NULL AT THE END
    if(!path) {
        FSFree(pathd);
        return NULL;
    }

    // NOW COPY ALL THE NAMES
    totalname=0;
    for(idx=0;idx<depth;++idx)
    {
        if(pathd[idx].string) FSMemcpy(path+totalname,pathd[idx].string,pathd[idx].length);
        totalname+=pathd[idx].length;
        path[totalname]='/';
        ++totalname;
    }
    if(pathd[idx].string) FSMemcpy(path+totalname,pathd[idx].string,pathd[idx].length);
    totalname+=pathd[idx].length;
    path[totalname]=0;
    FSFree(pathd);
    return path;
}


// REMOVE AN EMPTY DIRECTORY

int FSRmdir(clfs_char *name,int force,struct FS_VOLUME *fs)
{
    int error;
    struct FS_FILE *dir,*entry;


     error=FSOpenDir(name,&dir,fs);

     if(error!=FS_OK) {
         return error;
     }

     if(!FSFileIsOpen(dir,fs)) {
        // THIS INDICATES THAT THE DIRECTORY WAS OPEN EXCLUSIVELY BY OUR CALL TO FSOpenDir()
        // NOT NECESSARILY OBVIOUS, SINCE FSOpenDir MIGHT RETURN A POINTER TO AN EXISTING DIRECTORY BEING USED BY AN OPEN FILE
     return FS_USED;   // BEING USED BY OTHER FILES OR IT'S CURRENT DIR OR ROOT DIR. CAN'T DELETE.
     }

     if(force) {

            // RECURSIVELY REMOVE ALL DIRECTORIES WITHIN!

     FSSeek(dir,0,SEEK_SET);

     while( (error=FSGetNextDirEntry(dir,&entry))==FS_OK)
     {
         if(entry->F.EntryType==DE_FILE) {
         if(entry->F.Attributes&FSATTR_DIR)
         {
             if(FSNameCompare(entry->Name,".",FS_FALSE))
             {
                     // SAVE FOR LAST MINUTE IN CASE WE NEED TO QUIT UNEXPECTEDLY
                     FSFreeFile(entry);
                     continue;
             }

          // RECURSE SUBDIRECTORIES
             clfs_char *temp=FSGetFullPath(entry);
             int error2=FSRmdir(temp,FS_TRUE,fs);
             FSFree(temp);
             if(error2!=FS_OK) return error2;



         }
         else {
            // JUST A FILE, GO AHEAD AND DELETE IT
            entry->F.EntryType=DE_DELETED;
            FSFreeAllClusters(entry);
            FSTouchDirEntry(entry);

         }
         }

         FSFreeFile(entry);
     }
     if(error!=FS_EOF) { FSClose(dir); return error; }
     // NOW REMOVE THE ACTUAL DIRECTORY
     dir->F.EntryType=DE_DELETED;
     error=FSTouchDirEntry(dir);
     dir->F.EntryType=DE_FILE;
     FSClose(dir);
     return error;
     }
     else {
         // DO NOT REMOVE IF DIRECTORY IS NOT EMPTY
         FSSeek(dir,0,SEEK_SET);

         while( (error=FSGetNextDirEntry(dir,&entry))==FS_OK)
         {
             if(entry->F.EntryType==DE_FILE) {
             if(entry->F.Attributes&FSATTR_DIR)
             {
                 if(!FSNameCompare(entry->Name,".",FS_FALSE)) break;
             }
             else break;
             }
             FSFreeFile(entry);
         }
         if(error==FS_OK) {
             FSFreeFile(entry);
             FSClose(dir);
             return FS_USED;
         }
         // NOW REMOVE THE ACTUAL DIRECTORY
         dir->F.EntryType=DE_DELETED;
         error=FSTouchDirEntry(dir);
         dir->F.EntryType=DE_FILE;
         FSClose(dir);
         return error;
     }
}

